#!/usr/bin/env python

# Standard library imports
import argparse
import sys

import numpy as np
import pandas as pd
import scipy.stats as stats

from sklearn_genetic.space import Continuous, Categorical, Integer

# from pgsui import GenotypeData
from snpio import GenotypeData
from impute.estimators import (
    ImputeNLPCA,
    ImputeUBP,
    ImputeRandomForest,
    ImputeVAE,
)
from impute.simple_imputers import ImputePhylo

# from snpio import GenotypeData
# from impute.estimators import *
# from impute.simple_imputers import ImputeAlleleFreq, ImputePhylo

# from read_input import GenotypeData
# from estimators import *


def main():
    """Class instantiations and main package body"""

    args = get_arguments()

    if args.str and args.phylip:
        sys.exit("Error: Only one file type can be specified")

        # If VCF file is specified.
    if args.str:
        if not args.pop_ids and args.popmap is None:
            raise TypeError("Either --pop_ids or --popmap must be specified\n")

        if args.pop_ids:
            print("\n--pop_ids was specified as column 2\n")
        else:
            print(
                "\n--pop_ids was not specified; "
                "using popmap file to get population IDs\n"
            )

        if args.onerow_perind:
            print("\nUsing one row per individual...\n")
        else:
            print("\nUsing two rows per individual...\n")

        if args.onerow_perind:
            data = GenotypeData(
                filename=args.str,
                filetype="structure1row",
                popmapfile=args.popmap,
                guidetree=args.treefile,
                qmatrix_iqtree=args.iqtree,
            )
        else:
            data = GenotypeData(
                filename=args.str,
                filetype="structure2row",
                popmapfile=args.popmap,
                guidetree=args.treefile,
                qmatrix_iqtree=args.iqtree,
            )

    if args.phylip:
        if args.pop_ids or args.onerow_perind:
            print(
                "\nPhylip file was used with structure arguments; ignoring "
                "structure file arguments\n"
            )

        if args.popmap is None:
            raise TypeError("No popmap file supplied with PHYLIP file\n")

        data = GenotypeData(
            filename=args.phylip,
            filetype="phylip",
            popmapfile=args.popmap,
            guidetree=args.treefile,
            qmatrix_iqtree=args.iqtree,
            siterates_iqtree="pgsui/example_data/trees/test_n10.rate",
        )

    if args.resume_imputed:
        pass
        # data.read_imputed(args.resume_imputed, impute_methods="rf")
        # data.write_imputed(data.imputed_rf_df, args.prefix)

    else:
        # For randomizedsearchcv
        # Number of trees in random forest
        n_estimators = [
            int(x) for x in np.linspace(start=100, stop=1000, num=10)
        ]

        # Number of features to consider at every split
        max_features = ["sqrt", "log2"]

        # Maximum number of levels in the tree
        max_depth = [int(x) for x in np.linspace(10, 110, num=11)]
        max_depth.append(None)

        # Minimmum number of samples required to split a node
        min_samples_split = [int(x) for x in np.linspace(2, 10, num=5)]

        # Minimum number of samples required at each leaf node
        min_samples_leaf = [int(x) for x in np.linspace(1, 5, num=5)]

        # Proportion of dataset to use with bootstrapping
        # max_samples = [x for x in np.linspace(0.5, 1.0, num=6)]

        # # Random Forest gridparams - RandomizedSearchCV
        # grid_params = {
        #     "max_features": max_features,
        #     "max_depth": max_depth,
        #     "min_samples_split": min_samples_split,
        #     "min_samples_leaf": min_samples_leaf,
        # }

        # Random Forest gridparams - Genetic Algorithms
        # grid_params = {
        # 	"n_estimators": Integer(100, 500),
        # 	"max_features": max_features,
        # 	"max_depth": max_depth,
        # 	"min_samples_split": min_samples_split,
        # 	"min_samples_leaf": min_samples_leaf,
        # 	"max_samples": max_samples
        # }

        # # Genetic Algorithm grid_params
        # grid_params = {
        #     "max_features": Categorical(["sqrt", "log2"]),
        #     "min_samples_split": Integer(2, 10),
        #     "min_samples_leaf": Integer(1, 10),
        #     "max_depth": Integer(2, 110),
        # }

        # Bayesian Ridge gridparams - RandomizedSearchCV
        # grid_params = {
        # 	"alpha_1": stats.loguniform(1e-6, 0.01),
        # 	"alpha_2": stats.loguniform(1e-6, 0.01),
        # 	"lambda_1": stats.loguniform(1e-6, 0.01),
        # 	"lambda_2": stats.loguniform(1e-6, 0.01),
        # }

        # # Bayesian Ridge gridparams - Genetic algorithm
        # grid_params = {
        # 	"alpha_1": Continuous(1e-6, 1e-3, distribution="log-uniform"),
        # 	"alpha_2": Continuous(1e-6, 1e-3, distribution="log-uniform"),
        # 	"lambda_1": Continuous(1e-6, 1e-3, distribution="log-uniform"),
        # 	"lambda_2": Continuous(1e-6, 1e-3, distribution="log-uniform")
        # }

        # # Random forest imputation with genetic algorithm grid search
        # rf_imp = ImputeRandomForest(
        #     data,
        #     prefix=args.prefix,
        #     n_estimators=50,
        #     n_nearest_features=1,
        #     gridparams=grid_params,
        #     cv=3,
        #     grid_iter=40,
        #     n_jobs=4,
        #     max_iter=2,
        #     column_subset=1.0,
        #     ga=False,
        #     disable_progressbar=True,
        #     extratrees=False,
        #     mutation_probability=0.1,
        #     progress_update_percent=20,
        #     chunk_size=1.0,
        #     initial_strategy="phylogeny",
        # )

        # # Genetic Algorithm grid search Test
        # rf_imp2 = ImputeRandomForest(
        #     data,
        #     prefix=args.prefix,
        #     n_estimators=50,
        #     n_nearest_features=2,
        #     gridparams=grid_params,
        #     cv=3,
        #     grid_iter=40,
        #     n_jobs=-1,
        #     max_iter=2,
        #     column_subset=1.0,
        #     ga=True,
        #     disable_progressbar=True,
        #     extratrees=False,
        #     chunk_size=1.0,
        #     initial_strategy="phylogeny",
        # )

        # rfdata = rf_imp.imputed
        # print(rfdata.genotypes012_df)

        # rf_data = rf_imp.imputed
        # print(data.genotypes012_df)
        # print(rf_data.genotypes012_df)

        # imp_decoded = data.decode_imputed(rf_imp.imputed)
        # print(imp_decoded)

        # # RandomizedSearchCV Test
        # rf_imp = ImputeRandomForest(
        #     data,
        #     prefix=args.prefix,
        #     n_estimators=50,
        #     n_nearest_features=3,
        #     gridparams=grid_params,
        #     cv=3,
        #     grid_iter=40,
        #     n_jobs=4,
        #     max_iter=2,
        #     column_subset=5,
        #     ga=False,
        #     disable_progressbar=False,
        #     extratrees=False,
        #     progress_update_percent=20,
        #     chunk_size=0.2,
        #     initial_strategy="phylogeny",
        # )

        # lgbm = ImputeLightGBM(
        #     data,
        #     prefix=args.prefix,
        #     cv=3,
        #     n_jobs=4,
        #     n_estimators=50,
        #     disable_progressbar=True,
        #     chunk_size=0.2,
        #     validation_only=0.1,
        #     n_nearest_features=3,
        #     max_iter=2,
        #     initial_strategy="populations",
        # )

        # vae = ImputeVAE(
        #     genotype_data=data,
        #     prefix=args.prefix,
        #     disable_progressbar=True,
        #     validation_only=None,
        #     initial_strategy="populations",
        # )

        # vae_gtdata = vae.imputed
        # print(vae_gtdata.genotypes012_df)

        # complete_encoded = imputer.train(train_epochs=300, batch_size=256)
        # print(complete_encoded)

        # rf_imp = ImputeRandomForest(
        #     data,
        #     prefix=args.prefix,
        #     n_estimators=50,
        #     n_nearest_features=3,
        #     n_jobs=4,
        #     max_iter=2,
        #     disable_progressbar=True,
        #     extratrees=False,
        #     max_features="sqrt",
        #     min_samples_split=5,
        #     min_samples_leaf=2,
        #     max_depth=30,
        #     cv=3,
        #     validation_only=0.3,
        #     chunk_size=1.0,
        #     initial_strategy="populations",
        # )

        # afpops = ImputeAlleleFreq(
        #     genotype_data=data,
        #     by_populations=True,
        #     prefix=args.prefix,
        # )

        # print(data.genotypes012_df)
        # print(afpops.genotypes012_df)

        # br_imp = ImputeBayesianRidge(data, prefix=args.prefix, n_iter=100, gridparams=grid_params, grid_iter=3, cv=3, n_jobs=4, max_iter=5, n_nearest_features=3, column_subset=4, ga=False, disable_progressbar=True, progress_update_percent=20, chunk_size=1.0)

        # aftestpops = ImputeAlleleFreq(
        #     genotype_data=data, by_populations=True, prefix=args.prefix
        # )

        # aftestpops_data = aftestpops.imputed

        # print(data.genotypes012_df)
        # print(aftestpops_data.genotypes012_df)

    # vae = ImputeVAE(
    #     gt=np.array([[0, 1], [-9, 1], [2, -9]]),
    #     initial_strategy="most_frequent",
    #     cv=3,
    #     validation_only=None,
    # )

    # vae_data = vae.imputed

    # print(data.genotypes012_df)
    # print(vae_data.genotypes012_df)

    # For GridSearchCV. Generate parameters to sample from.
    learning_rate = [float(10) ** x for x in np.arange(-4, 0)]
    l1_penalty = [float(10) ** x for x in np.arange(-6, -1)]
    l1_penalty.append(0.0)
    l2_penalty = [float(10) ** x for x in np.arange(-6, -1)]
    l2_penalty.append(0.0)
    hidden_activation = ["elu", "relu"]
    num_hidden_layers = [1, 2, 3, 4, 5]
    hidden_layer_sizes = ["sqrt", "midpoint"]
    n_components = [2, 3]
    dropout_rate = [round(x, 1) for x in np.arange(0.0, 1.0, 0.1)]
    batch_size = [16, 32, 48, 64]
    optimizer = ["adam", "sgd", "adagrad"]

    # grid_params = {
    #     "learning_rate": Continuous(1e-6, 0.1, distribution="log-uniform"),
    #     "l2_penalty": Continuous(1e-6, 0.01, distribution="uniform"),
    #     "n_components": Integer(2, 3),
    #     # "hidden_activation": Categorical(["elu", "relu"]),
    # }

    grid_params = {
        # "learning_rate": learning_rate,
        # "l1_penalty": l1_penalty,
        "l2_penalty": l2_penalty,
        # "hidden_activation": hidden_activation,
        # "hidden_layer_sizes": hidden_layer_sizes,
        "n_components": n_components,
        # "dropout_rate": dropout_rate,
        # "batch_size": batch_size,
        # "optimizer": optimizer,
    }

    ubp = ImputeUBP(
        data,
        disable_progressbar=False,
        cv=3,
        column_subset=1.0,
        validation_split=0.0,
        learning_rate=0.1,
        num_hidden_layers=1,
        verbose=1,
        dropout_rate=0.2,
        hidden_activation="elu",
        batch_size=64,
        l1_penalty=1e-6,
        l2_penalty=1e-6,
        gridparams=grid_params,
        n_jobs=4,
        grid_iter=5,
        sim_strategy="nonrandom_weighted",
        sim_prop_missing=0.4,
        scoring_metric="precision_recall_macro",
        gridsearch_method="randomized_gridsearch",
        early_stop_gen=5,
        # sample_weights={0: 1.0, 1: 0.0, 2: 1.0},
        # sample_weights="auto",
    )

    # ubp = ImputeVAE(
    #     data,
    #     # gridparams=grid_params,
    #     # initial_strategy="populations",
    #     # disable_progressbar=True,
    #     # cv=3,
    #     # column_subset=1.0,
    #     # validation_size=0.3,
    #     # learning_rate=0.1,
    #     # num_hidden_layers=1,
    #     # verbose=1,
    #     # gridparams=grid_params,
    # )

    # nlpca_data = nlpca.imputed
    # print(nlpca_data.genotypes012_df)

    # print(data.genotypes012_df)
    # print(nlpca_data.genotypes012_df)

    # ubp = ImputeUBP(
    #     genotype_data=data,
    #     test_categorical=np.array([[0, 1], [-9, 1], [2, -9]]),
    # )

    # ubp = ImputeVAE(
    #     gt=np.array([[0, 1], [-9, 1], [2, -9]]),
    #     initial_strategy="most_frequent",
    # )

    # br_imp = ImputeBayesianRidge(
    #     data,
    #     prefix=args.prefix,
    #     alpha_1=0.0002689638465560243,
    #     alpha_2=0.0001473822173361299,
    #     lambda_1=0.0003281735206234651,
    #     lambda_2=0.00020767920087590963,
    #     n_iter=100,
    #     n_nearest_features=3,
    #     progress_update_percent=20,
    #     disable_progressbar=True,
    #     max_iter=2,
    #     cv=3,
    #     initial_strategy="group_mode",
    # )

    # phylo = ImputePhylo(
    #     genotype_data=data, save_plots=False, disable_progressbar=True
    # )

    # phylodata = phylo.imputed
    # print(phylodata.genotypes012_df)


def get_arguments():
    """[Parse command-line arguments. Imported with argparse]

    Returns:
        [argparse object]: [contains command-line arguments; accessed as method]
    """

    parser = argparse.ArgumentParser(
        description="Machine learning missing data imputation and species delimitation",
        add_help=False,
    )

    required_args = parser.add_argument_group("Required arguments")
    filetype_args = parser.add_argument_group(
        "File type arguments (choose only one)"
    )
    structure_args = parser.add_argument_group("Structure file arguments")
    optional_args = parser.add_argument_group("Optional arguments")

    # File Type arguments
    filetype_args.add_argument(
        "-s", "--str", type=str, required=False, help="Input structure file"
    )
    filetype_args.add_argument(
        "-p", "--phylip", type=str, required=False, help="Input phylip file"
    )

    filetype_args.add_argument(
        "-t",
        "--treefile",
        type=str,
        required=False,
        default=None,
        help="Newick-formatted treefile",
    )

    filetype_args.add_argument(
        "-i",
        "--iqtree",
        type=str,
        required=False,
        help=".iqtree output file containing Rate Matrix Q",
    )

    # Structure Arguments
    structure_args.add_argument(
        "--onerow_perind",
        default=False,
        action="store_true",
        help="Toggles on one row per individual option in structure file",
    )
    structure_args.add_argument(
        "--pop_ids",
        default=False,
        required=False,
        action="store_true",
        help="Toggles on population ID column (2nd col) in structure file",
    )

    ## Optional Arguments
    optional_args.add_argument(
        "-m",
        "--popmap",
        type=str,
        required=False,
        default=None,
        help="Two-column tab-separated population map file: inds\tpops. No header line",
    )
    optional_args.add_argument(
        "--prefix",
        type=str,
        required=False,
        default="output",
        help="Prefix for output files",
    )

    optional_args.add_argument(
        "--resume_imputed",
        type=str,
        required=False,
        help="Read in imputed data from a file instead of doing the imputation",
    )
    # Add help menu
    optional_args.add_argument(
        "-h", "--help", action="help", help="Displays this help menu"
    )

    # If no command-line arguments are called then exit and call help menu.
    if len(sys.argv) == 1:
        print("\nExiting because no command-line options were called.\n")
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    main()



# Standard library imports
import gc
import math
import os
import sys
import warnings

# from collections import namedtuple
from contextlib import redirect_stdout
from time import time
from typing import Optional, Union, List, Dict, Tuple, Any, Callable

# Third-party imports
## For stats and numeric operations
import numpy as np
import pandas as pd
from scipy import stats

# scikit-learn imports
from sklearn.base import clone
from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import IterativeImputer
from sklearn.impute import SimpleImputer
from sklearn.impute._base import _check_inputs_dtype

## For warnings
from sklearn.exceptions import ConvergenceWarning
from sklearn.utils._testing import ignore_warnings

## Required for IterativeImputer.fit_transform()
from sklearn.utils import check_random_state, _safe_indexing, is_scalar_nan
from sklearn.utils._mask import _get_mask
from sklearn.utils.validation import FLOAT_DTYPES
from sklearn.preprocessing import LabelEncoder

# Custom function imports
try:
    from .. import simple_imputers
    from ...utils.misc import get_processor_name
    from ...utils.misc import HiddenPrints
    from ...utils.misc import isnotebook
except (ModuleNotFoundError, ValueError, ImportError):
    from pgsui.impute import simple_imputers
    from pgsui.utils.misc import get_processor_name
    from pgsui.utils.misc import HiddenPrints
    from pgsui.utils.misc import isnotebook

# Uses scikit-learn-intellex package if CPU is Intel
if get_processor_name().strip().startswith("Intel"):
    try:
        from sklearnex import patch_sklearn

        patch_sklearn(verbose=False)
    except (ImportError, TypeError):
        print(
            "Processor not compatible with scikit-learn-intelex; using "
            "default configuration"
        )

is_notebook = isnotebook()

if is_notebook:
    from tqdm.notebook import tqdm as progressbar
else:
    if sys.platform == "linux" or sys.platform == "linux2":
        from tqdm.auto import tqdm as progressbar
    else:
        from tqdm import tqdm as progressbar

# NOTE: Removed ImputeTriplets to save memory.
# ImputerTriplet is there so that the missing values
# can be predicted on an already-fit model using just the
# transform method. I didn't need it, so I removed it
# because it was saving thousands of fit estimator models into the object

# _ImputerTripletAll = namedtuple(
# 	'_ImputerTripletAll', ['feat_idx', 'neighbor_feat_idx', 'estimator'])


class IterativeImputerFixedParams(IterativeImputer):
    """Overridden IterativeImputer methods.

    Herein, progress status updates, optimizations to save RAM, and several other improvements have been added. IterativeImputer is a multivariate imputer that estimates each feature from all the others. A strategy for imputing missing values by modeling each feature with missing values as a function of other features in a round-robin fashion.Read more in the scikit-learn User Guide for IterativeImputer. scikit-learn versionadded: 0.21. NOTE: This estimator is still **experimental** for now: the predictions and the API might change without any deprecation cycle. To use it, you need to explicitly import ``enable_iterative_imputer``\.

        IterativeImputer is based on the R MICE (Multivariate Imputation by Chained Equationspackage) [3]_. See [4]_ for more information about multiple versus single imputations.

        >>> # explicitly require this experimental feature
        >>> from sklearn.experimental import enable_iterative_imputer
        >>>
        >>> # now you can import normally from sklearn.impute
        >>> from sklearn.impute import IterativeImputer

    Args:
        logfilepath (str): Path to the progress log file.

        clf_kwargs (Dict[str, Any]): A dictionary with the classifier keyword arguments.

        prefix (str): Prefix for output files.

        estimator (callable estimator object, optional): The estimator to use at each step of the round-robin imputation. If ``sample_posterior`` is True, the estimator must support ``return_std`` in its ``predict`` method. Defaults to BayesianRidge().

        clf_type (str, optional): Whether to run ```'classifier'``` or ``'regression'`` based imputation. Defaults to 'classifier'

        disable_progressbar (bool, optional): Whether or not to disable the tqdm progress bar. If True, disables the progress bar. If False, tqdm is used for the progress bar. This can be useful if you are running the imputation on an HPC cluster or are saving the standard output to a file. If True, progress updates will be printed to the screen every ``progress_update_percent`` iterations. Defaults to False.

        progress_update_percent (int, optional): How often to display progress updates (as a percentage) if ``disable_progressbar`` is True. If ``progress_update_frequency=10``\, then it displays progress updates every 10%. Defaults to 10.

        pops (List[Union[str, int]] or None): List of population IDs to be used with ImputeAlleleFreq if ``initial_strategy="populations"``\.

        missing_values (int or np.nan, optional): The placeholder for the missing values. All occurrences of ``missing_values`` will be imputed. For pandas dataframes with	nullable integer dtypes with missing values, ``missing_values`` should be set to ``np.nan``\, since ``pd.NA`` will be converted to ``np.nan``\. Defaults to np.nan.

        Sample_posterior (bool, optional): Whether to sample from the (Gaussian) predictive posterior of the fitted estimator for each imputation. Estimator must support ``return_std`` in its ``predict`` method if set to ``True``\. Set to ``True`` if using ``IterativeImputer`` for multiple imputations. Defaults to False.

        max_iter (int, optional): Maximum number of imputation rounds to perform before returning the imputations computed during the final round. A round is a single	imputation of each feature with missing values. The stopping criterion is met once ``max(abs(X_t - X_{t-1}))/max(abs(X[known_vals])) < tol``\,	where ``X_t`` is ``X`` at iteration ``t``\. Note that early stopping is only applied if ``sample_posterior=False``\. Defaults to 10.

        tol (float, optional): Tolerance of the stopping condition. Defaults to 1e-3.

        n_nearest_features (int, optional): Number of other features to use to estimate the missing values of each feature column. Nearness between features is measured using the absolute correlation coefficient between each feature pair (after initial imputation). To ensure coverage of features throughout the imputation process, the neighbor features are not necessarily nearest,	but are drawn with probability proportional to correlation for each	imputed target feature. Can provide significant speed-up when the number of features is huge. If ``None``\, all features will be used. Defaults to None.

        initial_strategy (str, optional): Which strategy to use to initialize the missing values. Same as the ``strategy`` parameter in :class:`~sklearn.impute.SimpleImputer`	Valid values: "mean", "median", "most_frequent", "populations", "phylogeny", "mf", or "constant". Defaults to "mean".

        imputation_order (str, optional): The order in which the features will be imputed. Possible values: "ascending" (From features with fewest missing values to most), "descending" (From features with most missing values to fewest, "roman" (Left to right), "arabic" (Right to left),  random" (A random order for each round). Defaults to "ascending".

        skip_complete (bool, optional): If ``True`` then features with missing values during ``transform`` that did not have any missing values during ``fit`` will be imputed with the initial imputation method only. Set to ``True`` if you have	many features with no missing values at both ``fit`` and ``transform`` time to save compute. Defaults to False.

        min_value (float or array-like of shape (n_features,), optional): Minimum possible imputed value. Broadcast to shape (n_features,) if scalar. If array-like, expects shape (n_features,), one min value for each feature. The default is `-np.inf`...sklearn versionchanged:: 0.23 (Added support for array-like). Defaults to -np.inf.

        max_value (float or array-like of shape (n_features,), optional): Maximum possible imputed value. Broadcast to shape (n_features,) if scalar. If array-like, expects shape (n_features,), one max value for each feature..sklearn versionchanged:: 0.23 (Added support for array-like). Defaults to np.inf.

        verbose (int, optional): Verbosity flag, controls the debug messages that are issued as functions are evaluated. The higher, the more verbose. Can be 0, 1, or 2. Defaults to 0.

        random_state (int or RandomState instance, optional): The seed of the pseudo random number generator to use. Randomizes selection of estimator features if n_nearest_features is not None, the ``imputation_order`` if ``random``\, and the sampling from posterior if ``sample_posterior`` is True. Use an integer for determinism. Defaults to None.

        add_indicator (bool, optional): If True, a :class:`MissingIndicator` transform will stack onto output of the imputer's transform. This allows a predictive estimator to account for missingness despite imputation. If a feature has no missing values at fit/train time, the feature won't appear on the missing indicator even if there are missing values at transform/test time. Defaults to False.

        genotype_data (GenotypeData object, optional): GenotypeData object containing dictionary with keys=sampleIds and values=list of genotypes for the corresponding key. If using ``initial_strategy="phylogeny``\, then this object also needs contain the treefile and qmatrix objects. Defaults to None.

        str_encodings (dict(str: int), optional): Integer encodings used in STRUCTURE-formatted file. Should be a dictionary with keys=nucleotides and values=integer encodings. The missing data encoding should also be included. Argument is ignored if using a PHYLIP-formatted file. Defaults to {"A": 1, "C": 2, "G": 3, "T": 4, "N": -9}

        kwargs (Dict[str, Any]): For compatibility with grid search IterativeImputer.

    Attributes:
        initial_imputer_ (sklearn.impute.SimpleImputer):  Imputer used to initialize the missing values.

        n_iter_ (int): Number of iteration rounds that occurred. Will be less than ``self.max_iter`` if early stopping criterion was reached.

        n_features_with_missing_ (int): Number of features with missing values.

        indicator_ (sklearn.impute.MissingIndicator): Indicator used to add binary indicators for missing values ``None`` if add_indicator is False.

        random_state_ (RandomState instance): RandomState instance that is generated either from a seed, the random number generator or by ``np.random``\.

        logfilepath (str): Path to status logfile.

        clf_kwargs (Dict[str, Any]): Keyword arguments for estimator.

        prefix (str): Prefix for output files.

        clf_type (str): Type of estimator, either "classifier" or "regressor".

        disable_progressbar (bool): Whether to disable the tqdm progress bar. If True, writes status updates to file instead of tqdm progress bar.

        progress_update_percent (float or None): Print feature progress update every ``progress_update_percent`` percent.

        pops (List[Union[str, int]]): List of population IDs of shape (n_samples,).

        estimator (estimator object): Estimator to impute data with.

        sample_posterior (bool): Whether to use the sample_posterior option. This overridden class does not currently support sample_posterior.

        max_iter (int): The maximum number of iterations to run.

        tol (float): Convergence criteria.

        n_nearest_features (int): Number of nearest features to impute target with.

        initial_strategy (str): Strategy to use with SimpleImputer for training data.

        imputation_order (str): Order to impute.

        skip_complete (bool): Whether to skip features with no missing data.

        min_value (int or float): Minimum value of imputed data.

        max_value (int or float): Maximum value of imputed data.

        verbose (int): Verbosity level.

        genotype_data (GenotypeData object): GenotypeData object.

        str_encodings (Dict[str, int]): Dictionary with integer encodings for converting from STRUCTURE-formatted file to IUPAC nucleotides.

    See Also:
            SimpleImputer : Univariate imputation of missing values.

    Examples:
        >>> import numpy as np
        >>> from sklearn.experimental import enable_iterative_imputer
        >>> from sklearn.impute import IterativeImputer
        >>> imp_mean = IterativeImputer(random_state=0)
        >>> imp_mean.fit([[7, 2, 3], [4, np.nan, 6], [10, 5, 9]])
        >>> X = [[np.nan, 2, 3], [4, np.nan, 6], [10, np.nan, 9]]
        >>> imp_mean.transform(X)
        array([[ 6.9584...,  2.       ,  3.        ],
                        [ 4.       ,  2.6000...,  6.        ],
                        [10.       ,  4.9999...,  9.        ]])

    Notes:
        To support imputation in inductive mode we store each feature's estimator during the ``fit`` phase, and predict without refitting (in order) during	the ``transform`` phase. Features which contain all missing values at ``fit`` are discarded upon ``transform``\.

        NOTE: Inductive mode support was removed herein.

    References:
        .. [3] Stef van Buuren, Karin Groothuis-Oudshoorn (2011). mice: Multivariate Imputation by Chained Equations in R. Journal of Statistical Software 45: 1-67.

        .. [4] S. F. Buck, (1960). A Method of Estimation of Missing Values in	Multivariate Data Suitable for use with an Electronic Computer. Journal of the Royal Statistical Society 22(2): 302-306.
    """

    def __init__(
        self,
        logfilepath: str,
        clf_kwargs: Dict[str, Any],
        *,
        estimator: Callable = None,
        clf_type: str = "classifier",
        disable_progressbar: bool = False,
        progress_update_percent: Optional[int] = None,
        pops: Optional[List[Union[str, int]]] = None,
        missing_values: Union[float, int] = np.nan,
        sample_posterior: bool = False,
        max_iter: int = 10,
        tol: float = 1e-3,
        n_nearest_features: Optional[int] = None,
        initial_strategy: str = "mean",
        imputation_order: str = "ascending",
        skip_complete: bool = False,
        min_value: Union[int, float] = -np.inf,
        max_value: Union[int, float] = np.inf,
        verbose: int = 0,
        random_state: Optional[int] = None,
        add_indicator: bool = False,
        genotype_data: Optional[Any] = None,
        str_encodings: Optional[Dict[str, int]] = None,
        prefix="imputer",
        **kwargs,
    ) -> None:
        super().__init__(
            estimator=estimator,
            missing_values=missing_values,
            sample_posterior=sample_posterior,
            max_iter=max_iter,
            tol=tol,
            n_nearest_features=n_nearest_features,
            initial_strategy=initial_strategy,
            imputation_order=imputation_order,
            skip_complete=skip_complete,
            min_value=min_value,
            max_value=max_value,
            verbose=verbose,
            random_state=random_state,
            add_indicator=add_indicator,
        )

        self.logfilepath = logfilepath
        self.clf_kwargs = clf_kwargs
        self.prefix = prefix
        self.clf_type = clf_type
        self.disable_progressbar = disable_progressbar
        self.progress_update_percent = progress_update_percent
        self.pops = pops
        self.estimator = estimator
        self.sample_posterior = sample_posterior
        self.max_iter = max_iter
        self.tol = tol
        self.n_nearest_features = n_nearest_features
        self.initial_strategy = initial_strategy
        self.imputation_order = imputation_order
        self.skip_complete = skip_complete
        self.min_value = min_value
        self.max_value = max_value
        self.verbose = verbose
        self.random_state = random_state
        self.genotype_data = genotype_data
        self.str_encodings = str_encodings
        self.missing_values = missing_values

    def _initial_imputation(
        self, X: np.ndarray, cols_to_keep: np.ndarray, in_fit: bool = False
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Perform initial imputation for input X.

        Performs initial imputation on training data (neighbors).

        Args:
            X (ndarray): Input data of shape (n_samples, n_features), where n_samples is the number of samples and n_features is the number of features.

            cols_to_keep (numpy.ndarray): Column indices of shape (n_features,) to keep. Only used if ``initial_strategy=="phylogeny"``\.

            in_fit (bool, optional): True if function is called in fit, otherwise False. Defaults to False.

        Returns:
            Xt (numpy.ndarray): Input data of shape (n_samples, n_features), where n_samples is the number of samples and "n_features" is the number of features.

            X_filled (numpy.ndarray): Input data of shape (n_samples, features) with the most recent imputations.

            mask_missing_values (numpy.ndarray): Input data missing indicator matrix of of shape (n_samples, n_features), where n_samples is the number of samples and "n_features" is the number of features.

            X_missing_mask (numpy.ndarray): Input data mask matrix of shape (n_samples, n_features) indicating missing datapoints, where
                n_samples is the number of samples and n_features is the
                number of features.

        Raises:
            AttributeError: GenotypeData object must be initialized with guidetree and qmatrix if using ``initial_strategy=phylogeny``\.
        """
        if is_scalar_nan(self.missing_values):
            force_all_finite = "allow-nan"
        else:
            force_all_finite = True

        X = self._validate_data(
            X,
            dtype=FLOAT_DTYPES,
            order="F",
            reset=in_fit,
            force_all_finite=force_all_finite,
        )

        X[X < 0] = np.nan

        _check_inputs_dtype(X, self.missing_values)

        X_missing_mask = _get_mask(X, self.missing_values)
        mask_missing_values = X_missing_mask.copy()

        if self.initial_strategy == "populations":
            self.initial_imputer_ = simple_imputers.ImputeAlleleFreq(
                self.genotype_data,
                gt=np.nan_to_num(X, nan=-9).tolist(),
                pops=self.pops,
                by_populations=True,
                missing=-9,
                verbose=False,
                iterative_mode=True,
                validation_mode=True,
            )

            X_filled = np.array(self.initial_imputer_.imputed)
            Xt = X.copy()

        elif self.initial_strategy == "phylogeny":
            if (
                self.genotype_data.qmatrix is None
                and self.genotype_data.qmatrix_iqtree is None
            ) or self.genotype_data.guidetree is None:
                raise AttributeError(
                    "GenotypeData object was not initialized with "
                    "qmatrix/ qmatrix_iqtree or guidetree arguments, "
                    "but initial_strategy == 'phylogeny'"
                )

            else:
                self.initial_imputer_ = simple_imputers.ImputePhylo(
                    genotype_data=self.genotype_data,
                    str_encodings=self.str_encodings,
                    write_output=False,
                    disable_progressbar=True,
                    column_subset=cols_to_keep,
                    validation_mode=True,
                )

                X_filled = self.initial_imputer_.imputed.to_numpy()
                valid_sites = self.initial_imputer_.valid_sites

                valid_mask = np.flatnonzero(
                    np.logical_not(np.isnan(valid_sites))
                )

                Xt = X[:, valid_mask]
                mask_missing_values = mask_missing_values[:, valid_mask]

        elif self.initial_strategy == "mf":
            self.initial_imputer_ = simple_imputers.ImputeMF(
                self.genotype_data,
                gt=np.nan_to_num(X, nan=-9),
                missing=-9,
                verbose=False,
                validation_mode=True,
            )

            X_filled = np.array(self.initial_imputer_.imputed)
            Xt = X.copy()

        elif self.initial_strategy == "phylogeny":
            if (
                self.genotype_data.qmatrix is None
                and self.genotype_data.qmatrix_iqtree is None
            ) or self.genotype_data.guidetree is None:
                raise AttributeError(
                    "GenotypeData object was not initialized with "
                    "qmatrix/ qmatrix_iqtree or guidetree arguments, "
                    "but initial_strategy == 'phylogeny'"
                )

            else:
                self.initial_imputer_ = simple_imputers.ImputePhylo(
                    genotype_data=self.genotype_data,
                    str_encodings=self.str_encodings,
                    write_output=False,
                    disable_progressbar=True,
                    column_subset=cols_to_keep,
                    validation_mode=True,
                )

            X_filled = self.initial_imputer_.imputed.to_numpy()
            Xt = X.copy()

        else:
            if self.initial_imputer_ is None:
                self.initial_imputer_ = SimpleImputer(
                    missing_values=self.missing_values,
                    strategy=self.initial_strategy,
                )
                X_filled = self.initial_imputer_.fit_transform(X)

            else:
                X_filled = self.initial_imputer_.transform(X)

            valid_mask = np.flatnonzero(
                np.logical_not(np.isnan(self.initial_imputer_.statistics_))
            )

            Xt = X[:, valid_mask]
            mask_missing_values = mask_missing_values[:, valid_mask]

        return Xt, X_filled, mask_missing_values, X_missing_mask

    @ignore_warnings(category=UserWarning)
    def _impute_one_feature(
        self,
        X_filled: np.ndarray,
        mask_missing_values: np.ndarray,
        feat_idx: int,
        neighbor_feat_idx: np.ndarray,
        estimator: Optional[Any] = None,
        fit_mode: bool = True,
    ) -> np.ndarray:
        """Impute a single feature from the others provided.

        This function predicts the missing values of one of the features using the current estimates of all the other features. The ``estimator`` must support ``return_std=True`` in its ``predict`` method for this function to work.

        Args:
            X_filled (numpy.ndarray): Input data with the most recent imputations.

            mask_missing_values (numpy.ndarray): Input data's missing indicator matrix.

            feat_idx (int): Index of the feature currently being imputed.

            neighbor_feat_idx (numpy.ndarray): Indices of the features to be used in imputing ``feat_idx``\.

            estimator (object): The estimator to use at this step of the round-robin imputation If ``sample_posterior`` is True, the estimator must support ``return_std`` in its ``predict`` method.If None, it will be cloned from self._estimator.

            fit_mode (bool, optional): Whether to fit and predict with the estimator or just predict. Defaults to True.

        Returns:
            X_filled (numpy.ndarray)]: Input data with ``X_filled[missing_row_mask, feat_idx]`` updated.
        """
        if estimator is None and fit_mode is False:
            raise ValueError(
                "If fit_mode is False, then an already-fitted "
                "estimator should be passed in."
            )

        if estimator is None:
            estimator = clone(self._estimator)

        missing_row_mask = mask_missing_values[:, feat_idx]

        if fit_mode:
            X_train = _safe_indexing(
                X_filled[:, neighbor_feat_idx], ~missing_row_mask
            )

            y_train = _safe_indexing(X_filled[:, feat_idx], ~missing_row_mask)

            try:
                estimator.fit(X_train, y_train)
                le = None
            except ValueError as e:
                # Happens in newer versions of XGBClassifier.
                if str(e).startswith(
                    "Invalid classes inferred from unique values of `y`"
                ):
                    le = LabelEncoder()
                    y_train = le.fit_transform(y_train)
                    estimator.fit(X_train, y_train)

        # if no missing values, don't predict
        if np.sum(missing_row_mask) == 0:
            return X_filled

        X_test = _safe_indexing(
            X_filled[:, neighbor_feat_idx], missing_row_mask
        )

        if self.sample_posterior:
            raise NotImplementedError(
                "sample_posterior is not currently supported. "
                "Please set sample_posterior to False"
            )

        else:
            imputed_values = estimator.predict(X_test)

            if le is not None:
                imputed_values = le.inverse_transform(imputed_values)

            imputed_values = np.clip(
                imputed_values,
                self._min_value[feat_idx],
                self._max_value[feat_idx],
            )

        # update the feature
        X_filled[missing_row_mask, feat_idx] = imputed_values

        del estimator
        del X_train
        del y_train
        del X_test
        del imputed_values
        gc.collect()

        return X_filled

    @ignore_warnings(category=UserWarning)
    def fit_transform(
        self,
        X: np.ndarray,
        valid_cols: Optional[np.ndarray] = None,
        y: None = None,
    ) -> np.ndarray:
        """Fits the imputer on X and return the transformed X.

        Args:
            X (array-like, shape (n_samples, n_features)): Input data, where "n_samples" is the number of samples and "n_features" is the number of features.

            valid_cols (numpy.ndarray, optional): Array with column indices to keep. Defaults to None.

            y (None): Ignored. Here for compatibility with other sklearn classes.

        Returns:
            Xt (array-like, shape (n_samples, n_features)): The imputed input data.

        Raises:
            ValueError: "max_iter" must be a positive integer.
            ValueError: "tol" should be non-negative float.
            ValueError: One or more features has min_value >= max_value.
            ConvergenceWarning: Early stopping criterion not reached.
        """
        self.random_state_ = getattr(
            self, "random_state_", check_random_state(self.random_state)
        )

        if self.max_iter < 0:
            raise ValueError(
                f"'max_iter' should be a positive integer. Got {self.max_iter} instead."
            )

        if self.tol < 0:
            raise ValueError(
                f"'tol' should be a non-negative float. Got {self.tol} instead"
            )

        self._estimator = clone(self.estimator)

        self.initial_imputer_ = None

        # X is the input data subset to only valid features (not nan)
        # Xt is the data imputed with SimpleImputer
        # mask_missing_values is the missing indicator matrix
        # complete_mask is the input data's mask matrix
        X, Xt, mask_missing_values, complete_mask = self._initial_imputation(
            X, valid_cols, in_fit=True
        )

        super(IterativeImputer, self)._fit_indicator(complete_mask)
        X_indicator = super(IterativeImputer, self)._transform_indicator(
            complete_mask
        )

        if self.max_iter == 0 or np.all(mask_missing_values):
            self.n_iter_ = 0
            return super(IterativeImputer, self)._concatenate_indicator(
                Xt, X_indicator
            )

        # Edge case: a single feature. We return the initial ...
        if Xt.shape[1] == 1:
            self.n_iter_ = 0
            return super(IterativeImputer, self)._concatenate_indicator(
                Xt, X_indicator
            )

        self._min_value = self._validate_limit(
            self.min_value, "min", X.shape[1]
        )
        self._max_value = self._validate_limit(
            self.max_value, "max", X.shape[1]
        )

        if not np.all(np.greater(self._max_value, self._min_value)):
            raise ValueError(
                "One (or more) features have min_value >= max_value."
            )

        # order in which to impute
        # note this is probably too slow for large feature data (d > 100000)
        # and a better way would be good.
        # see: https://goo.gl/KyCNwj and subsequent comments
        ordered_idx = self._get_ordered_idx(mask_missing_values)
        total_features = len(ordered_idx)

        self.n_features_with_missing_ = len(ordered_idx)

        abs_corr_mat = self._get_abs_corr_mat(Xt)

        n_samples, n_features = Xt.shape

        if self.verbose > 0:
            print(
                f"[IterativeImputer] Completing matrix with shape ({X.shape},)"
            )
        start_t = time()

        if not self.sample_posterior:
            Xt_previous = Xt.copy()
            normalized_tol = self.tol * np.max(np.abs(X[~mask_missing_values]))

        total_iter = self.max_iter

        ###########################################
        ### Iteration Start
        ###########################################
        for self.n_iter_ in progressbar(
            range(1, total_iter + 1),
            desc="Iteration: ",
            disable=self.disable_progressbar,
        ):
            if self.imputation_order == "random":
                ordered_idx = self._get_ordered_idx(mask_missing_values)

            if self.disable_progressbar:
                with open(self.logfilepath, "a") as fout:
                    # Redirect to progress logfile
                    with redirect_stdout(fout):
                        print(
                            f"Iteration Progress: "
                            f"{self.n_iter_}/{self.max_iter} "
                            f"({int((self.n_iter_ / total_iter) * 100)}%)"
                        )

                if self.progress_update_percent is not None:
                    print_perc_interval = self.progress_update_percent

            ##########################
            ### Feature Start
            ##########################
            for i, feat_idx in enumerate(
                progressbar(
                    ordered_idx,
                    desc="Feature: ",
                    leave=False,
                    position=1,
                    disable=self.disable_progressbar,
                ),
                start=1,
            ):
                neighbor_feat_idx = self._get_neighbor_feat_idx(
                    n_features, feat_idx, abs_corr_mat
                )

                Xt = self._impute_one_feature(
                    Xt,
                    mask_missing_values,
                    feat_idx,
                    neighbor_feat_idx,
                    estimator=None,
                    fit_mode=True,
                )

                # NOTE: The below source code has been commented out to save
                # RAM. estimator_triplet object contains numerous fit estimators
                # that demand a lot of resources. It is primarily used for the
                # transform function, which is not needed in this application.

                # estimator_triplet = _ImputerTripletAll(
                # 	feat_idx,
                # 	neighbor_feat_idx,
                # 	estimator)

                # self.imputation_sequence_.append(estimator_triplet)

                # Only print feature updates at each progress_update_percent
                # interval
                if (
                    self.progress_update_percent is not None
                    and self.disable_progressbar
                ):
                    current_perc = math.ceil((i / total_features) * 100)

                    if current_perc >= print_perc_interval:
                        with open(self.logfilepath, "a") as fout:
                            # Redirect progress to file
                            with redirect_stdout(fout):
                                print(
                                    f"Feature Progress (Iteration "
                                    f"{self.n_iter_}/{self.max_iter}): "
                                    f"{i}/{total_features} ({current_perc}"
                                    f"%)"
                                )

                                if i == len(ordered_idx):
                                    print("\n", end="")

                        while print_perc_interval <= current_perc:
                            print_perc_interval += self.progress_update_percent

            if self.verbose > 1:
                print(
                    f"[IterativeImputer] Ending imputation round "
                    f"{self.n_iter_}/{self.max_iter}, "
                    f"elapsed time {(time() - start_t):0.2f}"
                )

            if not self.sample_posterior:
                inf_norm = np.linalg.norm(
                    Xt - Xt_previous, ord=np.inf, axis=None
                )

                if self.verbose > 0:
                    print(
                        f"[IterativeImputer] Change: {inf_norm}, "
                        f"scaled tolerance: {normalized_tol} "
                    )

                if inf_norm < normalized_tol:
                    if self.disable_progressbar:
                        # Early stopping criteria has been reached
                        with open(self.logfilepath, "a") as fout:
                            # Redirect to progress logfile
                            with redirect_stdout(fout):
                                print(
                                    "[IterativeImputer] Early stopping "
                                    "criterion reached."
                                )
                    else:
                        print(
                            "[IterativeImputer] Early stopping criterion "
                            "reached."
                        )

                    break
                Xt_previous = Xt.copy()

        else:
            if not self.sample_posterior:
                warnings.warn(
                    "[IterativeImputer] Early stopping criterion not"
                    " reached.",
                    ConvergenceWarning,
                )

        Xt[~mask_missing_values] = X[~mask_missing_values]
        Xt = Xt.astype(int)

        return (
            super(IterativeImputer, self)._concatenate_indicator(
                Xt, X_indicator
            ),
            None,
            None,
        )



# Standard Library Imports
import logging
import os
import pprint
import sys
import warnings

# Third-party Imports
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

# Grid search imports
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV

# Scikit-learn imports
from sklearn.base import BaseEstimator, TransformerMixin

# Genetic algorithm grid search imports
from sklearn_genetic import GASearchCV
from sklearn_genetic.callbacks import ConsecutiveStopping, ProgressBar
from sklearn_genetic.plots import plot_fitness_evolution

# Import tensorflow with reduced warnings.
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
logging.getLogger("tensorflow").disabled = True
warnings.filterwarnings("ignore", category=UserWarning)

# noinspection PyPackageRequirements
import tensorflow as tf

# Disable can't find cuda .dll errors. Also turns of GPU support.
tf.config.set_visible_devices([], "GPU")

from tensorflow.python.util import deprecation

# Disable warnings and info logs.
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
tf.get_logger().setLevel(logging.ERROR)


# Monkey patching deprecation utils to supress warnings.
# noinspection PyUnusedLocal
def deprecated(
    date, instructions, warn_once=True
):  # pylint: disable=unused-argument
    def deprecated_wrapper(func):
        return func

    return deprecated_wrapper


deprecation.deprecated = deprecated

from tensorflow.keras.callbacks import (
    ReduceLROnPlateau,
    CSVLogger,
)

# For development purposes
# from memory_profiler import memory_usage

# Custom module imports
try:
    from ...utils.misc import timer
    from ...utils.misc import isnotebook
    from ...utils.misc import validate_input_type
    from .neural_network_methods import NeuralNetworkMethods, DisabledCV
    from ...utils.scorers import Scorers
    from ...utils.plotting import Plotting
    from .callbacks import (
        UBPCallbacks,
        VAECallbacks,
        CyclicalAnnealingCallback,
    )
    from .keras_classifiers import VAEClassifier, MLPClassifier, SAEClassifier
    from ...data_processing.transformers import (
        SimGenotypeDataTransformer,
        AutoEncoderFeatureTransformer,
    )
except (ModuleNotFoundError, ValueError, ImportError):
    from utils.misc import timer
    from utils.misc import isnotebook
    from utils.misc import validate_input_type
    from impute.unsupervised.neural_network_methods import (
        NeuralNetworkMethods,
        DisabledCV,
    )
    from utils.scorers import Scorers
    from utils.plotting import Plotting
    from impute.unsupervised.callbacks import (
        UBPCallbacks,
        VAECallbacks,
        CyclicalAnnealingCallback,
    )
    from impute.unsupervised.keras_classifiers import (
        VAEClassifier,
        MLPClassifier,
        SAEClassifier,
    )
    from data_processing.transformers import (
        SimGenotypeDataTransformer,
        AutoEncoderFeatureTransformer,
    )

is_notebook = isnotebook()

if is_notebook:
    from tqdm.notebook import tqdm as progressbar
else:
    from tqdm import tqdm as progressbar

from tqdm.keras import TqdmCallback


class BaseNNImputer(BaseEstimator, TransformerMixin):
    """Base transformer class for neural network imputers.

    Args:
        genotype_data (GenotypeData): Input GenotypeData instance.

        prefix (str, optional): Prefix for output files. Defaults to "output".

        gridparams (Dict[str, Any] or None, optional): Dictionary with keys=keyword arguments for the specified estimator and values=lists of parameter values or distributions. If using GridSearchCV, distributions can be specified by using scipy.stats.uniform(low, high) (for a uniform distribution) or scipy.stats.loguniform(low, high) (useful if range of values spans orders of magnitude). ``gridparams`` will be used for a randomized grid search with cross-validation. If using the genetic algorithm grid search (GASearchCV) by setting ``ga=True``\, the parameters can be specified as ``sklearn_genetic.space`` objects. The grid search will determine the optimal parameters as those that maximize accuracy (or minimize root mean squared error for BayesianRidge regressor). NOTE: Takes a long time, so run it with a small subset of the data just to find the optimal parameters for the classifier, then run a full imputation using the optimal parameters. If ``gridparams=None``\, a grid search is not performed. Defaults to None.

        disable_progressbar (bool, optional): Whether to disable the tqdm progress bar. Useful if you are doing the imputation on e.g. a high-performance computing cluster, where sometimes tqdm does not work correctly. If False, uses tqdm progress bar. If True, does not use tqdm. Defaults to False.

        batch_size (int, optional): Batch size per epoch to train the model with.

        n_components (int, optional): Number of components to use as the input data. Defaults to 3.

        early_stop_gen (int, optional): Early stopping criterion for epochs. Training will stop if the loss (error) does not decrease past the tolerance level for ``early_stop_gen`` epochs. Will save the optimal model and reload it once ``early_stop_gen`` has been reached. Defaults to 25.

        num_hidden_layers (int, optional): Number of hidden layers to use in the model. Adjust if overfitting occurs. Defaults to 3.

        hidden_layer_sizes (str, List[int], List[str], or int, optional): Number of neurons to use in hidden layers. If string or a list of strings is supplied, the strings must be either "midpoint", "sqrt", or "log2". "midpoint" will calculate the midpoint as ``(n_features + n_components) / 2``. If "sqrt" is supplied, the square root of the number of features will be used to calculate the output units. If "log2" is supplied, the units will be calculated as ``log2(n_features)``. hidden_layer_sizes will calculate and set the number of output units for each hidden layer. If one string or integer is supplied, the model will use the same number of output units for each hidden layer. If a list of integers or strings is supplied, the model will use the values supplied in the list, which can differ. The list length must be equal to the ``num_hidden_layers``. Defaults to "midpoint".

        optimizer (str, optional): The optimizer to use with gradient descent. Possible value include: "adam", "sgd", "adagrad", "adadelta", "adamax", "ftrl", "nadam", and "rmsprop" are supported. See tf.keras.optimizers for more info. Defaults to "adam".

        hidden_activation (str, optional): The activation function to use for the hidden layers. See tf.keras.activations for more info. Commonly used activation functions include "elu", "relu", and "sigmoid". Defaults to "elu".

        learning_rate (float, optional): The learning rate for the optimizers. Adjust if the loss is learning too slowly. Defaults to 0.1.

        lr_patience (int, optional): Number of epochs with no loss improvement to wait before reducing the learning rate.

        epochs (int, optional): Maximum number of epochs to run if the ``early_stop_gen`` criterion is not met.

        weights_initializer (str, optional): Initializer to use for the model weights. See tf.keras.initializers for more info. Defaults to "glorot_normal".

        l1_penalty (float, optional): L1 regularization penalty to apply to reduce overfitting. Defaults to 0.01.

        l2_penalty (float, optional): L2 regularization penalty to apply to reduce overfitting. Defaults to 0.01.

        dropout_rate (float, optional): Dropout rate during training to reduce overfitting. Must be a float between 0 and 1. Defaults to 0.2.

        recurrent_weight (float, optional): Recurrent weight to calculate predictions. Defaults to 0.5.

        sample_weights (str or Dict[int, float], optional): Whether to weight each genotype by its class frequency. If ``sample_weights='auto'`` then it automatically calculates sample weights based on genotype class frequencies per locus; for example, if there are a lot more 0s and fewer 2s, then it will balance out the classes by weighting each genotype accordingly. ``sample_weights`` can also be a dictionary with the genotypes (0, 1, and 2) as the keys and the weights as the keys. If ``sample_weights`` is anything else, then they are not calculated. Defaults to False.

        grid_iter (int, optional): Number of iterations for grid search. Defaults to 50.

        gridsearch_method (str, optional): Grid search method to use. Possible options include: 'gridsearch', 'randomized_gridsearch', and 'genetic_algorithm'. 'gridsearch' runs all possible permutations of parameters, 'randomized_gridsearch' runs a random subset of parameters, and 'genetic_algorithm' uses a genetic algorithm gridsearch (via GASearchCV). Defaults to 'gridsearch'.

        ga_kwargs (Dict[str, Any] or None): Keyword arguments to be passed to a Genetic Algorithm grid search. Only used if ``ga==True``\.

        scoring_metric (str, optional): Scoring metric to use for randomized or genetic algorithm grid searches. See https://scikit-learn.org/stable/modules/model_evaluation.html for supported options. Defaults to "accuracy".

        sim_strategy (str, optional): Strategy to use for simulating missing data. Only used to validate the accuracy of the imputation. The final model will be trained with the non-simulated dataset. Supported options include: "random", "nonrandom", and "nonrandom_weighted". "random" randomly simulates missing data. When set to "nonrandom", branches from ``GenotypeData.guidetree`` will be randomly sampled to generate missing data on descendant nodes. For "nonrandom_weighted", missing data will be placed on nodes proportionally to their branch lengths (e.g., to generate data distributed as might be the case with mutation-disruption of RAD sites). Defaults to "random".

        sim_prop_missing (float, optional): Proportion of missing data to simulate with the SimGenotypeDataTransformer. Defaults to 0.1.

        n_jobs (int, optional): Number of parallel jobs to use in the grid search if ``gridparams`` is not None. -1 means use all available processors. Defaults to 1.

        verbose (int, optional): Verbosity setting. Can be 0, 1, or 2. 0 is the least and 2 is the most verbose. Defaults to 0.

        ToDo:
            Fix sample_weight for multi-label encodings.
    """

    def __init__(
        self,
        activate,
        nn_method,
        num_classes,
        act_func,
        *,
        genotype_data=None,
        prefix="imputer",
        gridparams=None,
        disable_progressbar=False,
        batch_size=32,
        n_components=3,
        early_stop_gen=25,
        num_hidden_layers=3,
        hidden_layer_sizes="midpoint",
        optimizer="adam",
        hidden_activation="elu",
        learning_rate=0.01,
        lr_patience=1,
        epochs=100,
        weights_initializer="glorot_normal",
        l1_penalty=0.0001,
        l2_penalty=0.0001,
        dropout_rate=0.2,
        sample_weights=False,
        grid_iter=80,
        gridsearch_method="gridsearch",
        ga_kwargs=None,
        scoring_metric="auc_macro",
        sim_strategy="random",
        sim_prop_missing=0.2,
        n_jobs=1,
        verbose=0,
        kl_beta=tf.Variable(1.0, trainable=False),
        validation_split=0.0,
        nlpca=False,
        testing=False,
    ):
        self.activate = activate
        self.act_func_ = act_func
        self.num_classes = num_classes
        self.testing = testing
        self.nn_method_ = nn_method

        self.genotype_data = genotype_data
        self.prefix = prefix
        self.gridparams = gridparams
        self.disable_progressbar = disable_progressbar
        self.batch_size = batch_size
        self.n_components = n_components

        self.early_stop_gen = early_stop_gen
        self.num_hidden_layers = num_hidden_layers
        self.hidden_layer_sizes = hidden_layer_sizes
        self.optimizer = optimizer
        self.hidden_activation = hidden_activation
        self.learning_rate = learning_rate
        self.lr_patience = lr_patience
        self.epochs = epochs
        self.weights_initializer = weights_initializer
        self.l1_penalty = l1_penalty
        self.l2_penalty = l2_penalty
        self.dropout_rate = dropout_rate
        self.sample_weights = sample_weights
        self.grid_iter = grid_iter
        self.gridsearch_method = gridsearch_method
        self.ga_kwargs = ga_kwargs
        self.scoring_metric = scoring_metric
        self.sim_strategy = sim_strategy
        self.sim_prop_missing = sim_prop_missing
        self.n_jobs = n_jobs
        self.verbose = verbose

        self.kl_beta = kl_beta
        self.validation_split = validation_split
        self.nlpca = nlpca

        self.run_gridsearch_ = False if self.gridparams is None else True
        self.is_multiclass_ = True if self.num_classes != 4 else False

        # Simulate missing data and get missing masks.
        self.sim = SimGenotypeDataTransformer(
            self.genotype_data,
            prop_missing=self.sim_prop_missing,
            strategy=self.sim_strategy,
            mask_missing=True,
        )

        # Binary encode y to get y_train.
        self.tt_ = AutoEncoderFeatureTransformer(
            num_classes=self.num_classes, activate=self.activate
        )

    @timer
    def fit(self, X):
        """Train the VAE model on input data X.

        Args:
            X (pandas.DataFrame, numpy.ndarray, or List[List[int]]): Input 012-encoded genotypes.

        Returns:
            self: Current instance; allows method chaining.

        Raises:
            TypeError: Must be either pandas.DataFrame, numpy.ndarray, or List[List[int]].
        """
        # Treating y as X here for compatibility with UBP/NLPCA.
        # With VAE, y=X anyways.
        y = X
        y = validate_input_type(y, return_type="array")

        self.nn_ = NeuralNetworkMethods()
        plotting = Plotting()

        if self.gridsearch_method == "genetic_algorithm":
            self.ga_ = True
        else:
            self.ga_ = False

        self.y_original_ = y.copy()
        self.y_simulated_ = self.sim.fit_transform(self.y_original_)

        # Get values where original value was not missing but missing data
        # was simulated.
        self.sim_missing_mask_ = self.sim.sim_missing_mask_

        # Original missing data.
        self.original_missing_mask_ = self.sim.original_missing_mask_

        # Both simulated and original missing data.
        self.all_missing_ = self.sim.all_missing_mask_

        # Just y_original with missing values encoded as -1.
        y_train = self.tt_.fit_transform(self.y_original_)

        if self.gridparams is not None:
            self.scoring_metrics_ = [
                "precision_recall_macro",
                "precision_recall_micro",
                "f1_score",
                "auc_macro",
                "auc_micro",
                "accuracy",
                "hamming",
            ]

        (
            logfile,
            callbacks,
            compile_params,
            model_params,
            fit_params,
        ) = self._initialize_parameters(y_train)

        if self.nn_method_ == "VAE":
            func = self.run_vae
        elif self.nn_method_ == "SAE":
            func = self.run_sae
        elif self.nn_method_ == "NLPCA":
            func = self.run_nlpca
        elif self.nn_method_ == "UBP":
            func = self.run_ubp
        else:
            raise ValueError(f"Invalid nn_method specified: {self.nn_method_}")

        (
            self.models_,
            self.histories_,
            self.best_params_,
            self.best_score_,
            self.best_estimator_,
            self.search_,
            self.metrics_,
        ) = func(
            self.y_original_,
            y_train,
            model_params,
            compile_params,
            fit_params,
        )

        if (
            self.best_params_ is not None
            and "optimizer__learning_rate" in self.best_params_
        ):
            self.best_params_["learning_rate"] = self.best_params_.pop(
                "optimizer__learning_rate"
            )

        if self.gridparams is not None:
            if self.verbose > 0:
                print("\nBest found parameters:")
                pprint.pprint(self.best_params_)
                print(f"\nBest score: {self.best_score_}")
            plotting.plot_grid_search(
                self.search_.cv_results_, self.nn_method_, self.prefix
            )

        plotting.plot_history(
            self.histories_, self.nn_method_, prefix=self.prefix
        )
        plotting.plot_metrics(
            self.metrics_, self.num_classes, self.prefix, self.nn_method_
        )

        if self.ga_:
            plot_fitness_evolution(self.search_)
            plt.savefig(
                os.path.join(
                    f"{self.prefix}_output",
                    "plots",
                    "Unsupervised",
                    self.nn_method_,
                    "fitness_evolution.pdf",
                ),
                bbox_inches="tight",
                facecolor="white",
            )
            plt.cla()
            plt.clf()
            plt.close()

            g = plotting.plot_search_space(self.search_)
            plt.savefig(
                os.path.join(
                    f"{self.prefix}_output",
                    "plots",
                    "Unsupervised",
                    self.nn_method_,
                    "search_space.pdf",
                ),
                bbox_inches="tight",
                facecolor="white",
            )
            plt.cla()
            plt.clf()
            plt.close()

        return self

    def transform(self, X):
        """Predict and decode imputations and return transformed array.

        Args:
            X (pandas.DataFrame, numpy.ndarray, or List[List[int]]): Input data to transform.

        Returns:
            numpy.ndarray: Imputed data.
        """
        y = X
        y = validate_input_type(y, return_type="array")

        if self.nn_method_ not in ["UBP", "NLPCA"]:
            model = self.models_[0]
        else:
            if len(self.models_) == 1:
                model = self.models_[0]
            else:
                model = self.models_[-1]

        y_true = y.copy()
        y_train = self.tt_.transform(y_true)
        y_true_1d = y_true.ravel()
        y_size = y_true.size
        y_missing_idx = np.flatnonzero(self.original_missing_mask_)

        if self.nn_method_ == "VAE":
            y_pred = model(
                tf.convert_to_tensor(y_train),
                training=False,
            )
        elif self.nn_method_ == "SAE":
            y_pred = model(y_train, training=False)
        else:
            y_pred = model(model.V_latent, training=False)
        y_pred = self.tt_.inverse_transform(y_pred)

        y_pred_decoded = self.nn_.decode_masked(
            y_train,
            y_pred,
            is_multiclass=self.is_multiclass_,
        )
        # y_pred_decoded, y_pred_certainty = self.nn_.decode_masked(
        #     y_train, y_pred, return_proba=True
        # )

        y_pred_1d = y_pred_decoded.ravel()

        # Only replace originally missing values at missing indexes.
        for i in np.arange(y_size):
            if i in y_missing_idx:
                y_true_1d[i] = y_pred_1d[i]

        self.nn_.write_gt_state_probs(
            y_pred,
            y_pred_1d,
            y_true,
            y_true_1d,
            self.nn_method_,
            self.sim_missing_mask_,
            self.original_missing_mask_,
            prefix=self.prefix,
        )

        Plotting.plot_confusion_matrix(
            y_true_1d, y_pred_1d, self.nn_method_, prefix=self.prefix
        )

        # if self.nn_method_ == "VAE":
        # Plotting.plot_label_clusters(z_mean, y_true_1d)

        # Return to original shape.
        return np.reshape(y_true_1d, y_true.shape)

    def run_clf(
        self,
        y_train,
        y_true,
        model_params,
        compile_params,
        fit_params,
        ubp_weights=None,
        phase=None,
        scoring=None,
        testing=False,
        **kwargs,
    ):
        """Run KerasClassifier with neural network model and grid search.

        Args:
            y_train (numpy.ndarray): Onehot-encoded training input data of shape (n_samples, n_features, num_classes).

            y_true (numpy.ndarray): Original 012-encoded input data of shape (n_samples, n_features).

            model_params (Dict[str, Any]): Dictionary with model parameters to be passed to KerasClassifier model.

            compile_params (Dict[str, Any]): Dictionary with params to be passed to Keras model.compile() in KerasClassifier.

            fit_params (Dict[str, Any]): Dictionary with parameters to be passed to fit in KerasClassifier.

            scoring (Dict[str, Callable], optional): Multimetric scorer made using sklearn.metrics.make_scorer. To be used with grid search.

        Returns:
            List[tf.keras.Model]: List of keras model objects. One for each phase (len=1 if NLPCA, len=3 if UBP).

            List[Dict[str, float]]: List of dictionaries with best neural network model history.

            Dict[str, Any] or None: Best parameters found during a grid search, or None if a grid search was not run.

            float: Best score obtained during grid search.

            tf.keras.Model: Best model found during grid search.

            sklearn.model_selection object (GridSearchCV, RandomizedSearchCV) or GASearchCV object.

            Dict[str, Any]: Per-class, micro, and macro-averaged metrics including accuracy, ROC-AUC, and Precision-Recall with Average Precision scores.
        """
        # This reduces memory usage.
        # tensorflow builds graphs that
        # will stack if not cleared before
        # building a new model.
        tf.keras.backend.clear_session()
        self.nn_.reset_seeds()

        model = None
        if self.nn_method_ in ["UBP", "NLPCA"]:
            V = model_params.pop("V")
            if phase is not None:
                desc = f"Epoch (Phase {phase}): "
            else:
                desc = "Epoch: "

        else:
            desc = "Epoch: "

        if not self.disable_progressbar and not self.run_gridsearch_:
            fit_params["callbacks"][-1] = TqdmCallback(
                epochs=self.epochs, verbose=self.verbose, desc=desc
            )

        if self.nn_method_ == "VAE":
            clf = VAEClassifier(
                **model_params,
                optimizer=compile_params["optimizer"],
                optimizer__learning_rate=compile_params["learning_rate"],
                loss=compile_params["loss"],
                metrics=compile_params["metrics"],
                run_eagerly=compile_params["run_eagerly"],
                callbacks=fit_params["callbacks"],
                epochs=fit_params["epochs"],
                verbose=0,
                num_classes=self.num_classes,
                activate=self.act_func_,
                fit__validation_split=fit_params["validation_split"],
                score__missing_mask=self.sim_missing_mask_,
                score__scoring_metric=self.scoring_metric,
                score__num_classes=self.num_classes,
                score__n_classes=self.num_classes,
            )
        elif self.nn_method_ == "SAE":
            clf = SAEClassifier(
                **model_params,
                optimizer=compile_params["optimizer"],
                optimizer__learning_rate=compile_params["learning_rate"],
                loss=compile_params["loss"],
                metrics=compile_params["metrics"],
                callbacks=fit_params["callbacks"],
                epochs=fit_params["epochs"],
                verbose=0,
                activate=self.act_func_,
                fit__validation_split=fit_params["validation_split"],
                score__missing_mask=self.sim_missing_mask_,
                score__scoring_metric=self.scoring_metric,
                score__num_classes=self.num_classes,
                score__n_classes=self.num_classes,
            )
        else:
            clf = MLPClassifier(
                V,
                y_train,
                **model_params,
                ubp_weights=ubp_weights,
                optimizer=compile_params["optimizer"],
                optimizer__learning_rate=compile_params["learning_rate"],
                loss=compile_params["loss"],
                metrics=compile_params["metrics"],
                epochs=fit_params["epochs"],
                phase=phase,
                callbacks=fit_params["callbacks"],
                validation_split=fit_params["validation_split"],
                verbose=0,
                score__missing_mask=self.sim_missing_mask_,
                score__scoring_metric=self.scoring_metric,
            )

        if self.run_gridsearch_:
            # Cannot do CV because there is no way to use test splits
            # given that the input gets refined. If using a test split,
            # then it would just be the randomly initialized values and
            # would not accurately represent the model.
            # Thus, we disable cross-validation for the grid searches.
            cross_val = DisabledCV()
            verbose = False if self.verbose == 0 else True

            if self.ga_:
                # Stop searching if GA sees no improvement.
                callback = [
                    ConsecutiveStopping(
                        generations=self.early_stop_gen, metric="fitness"
                    )
                ]

                if not self.disable_progressbar:
                    callback.append(ProgressBar())

                # Do genetic algorithm
                # with HiddenPrints():
                search = GASearchCV(
                    estimator=clf,
                    cv=cross_val,
                    scoring=scoring,
                    generations=self.grid_iter,
                    param_grid=self.gridparams,
                    n_jobs=self.n_jobs,
                    refit=self.scoring_metric,
                    verbose=verbose,
                    **self.ga_kwargs,
                    error_score="raise",
                )

                if self.nn_method_ in ["UBP", "NLPCA"]:
                    search.fit(V[self.n_components], y=y_true)
                else:
                    search.fit(y_true, y_true, callbacks=callback)

            else:
                # Write GridSearchCV to log file instead of STDOUT.
                if self.verbose >= 10:
                    old_stdout = sys.stdout
                    log_file = open(
                        os.path.join(
                            f"{self.prefix}_output",
                            "logs",
                            "Unsupervised",
                            self.nn_method_,
                            "gridsearch_progress_log.txt",
                        ),
                        "w",
                    )
                    sys.stdout = log_file

                if self.gridsearch_method.lower() == "gridsearch":
                    # Do GridSearchCV
                    search = GridSearchCV(
                        clf,
                        param_grid=self.gridparams,
                        n_jobs=self.n_jobs,
                        cv=cross_val,
                        scoring=scoring,
                        refit=self.scoring_metric,
                        verbose=self.verbose * 4,
                        error_score="raise",
                    )

                elif self.gridsearch_method.lower() == "randomized_gridsearch":
                    search = RandomizedSearchCV(
                        clf,
                        param_distributions=self.gridparams,
                        n_iter=self.grid_iter,
                        n_jobs=self.n_jobs,
                        cv=cross_val,
                        scoring=scoring,
                        refit=self.scoring_metric,
                        verbose=verbose * 4,
                        error_score="raise",
                    )

                else:
                    raise ValueError(
                        f"Invalid gridsearch_method specified: "
                        f"{self.gridsearch_method}"
                    )

                if self.nn_method_ in ["UBP", "NLPCA"]:
                    search.fit(V[self.n_components], y=y_true)
                else:
                    search.fit(y_true, y=y_true)

                if self.verbose >= 10:
                    # Make sure to revert STDOUT back to original.
                    sys.stdout = old_stdout
                    log_file.close()

            best_params = search.best_params_
            best_score = search.best_score_
            best_clf = search.best_estimator_

            fp = os.path.join(
                f"{self.prefix}_output",
                "reports",
                "Unsupervised",
                self.nn_method_,
                f"cvresults_{self.nn_method_}.csv",
            )

            cv_results = pd.DataFrame(search.cv_results_)
            cv_results.to_csv(fp, index=False)

        else:
            if self.nn_method_ in ["UBP", "NLPCA"]:
                clf.fit(V[self.n_components], y=y_true)
            else:
                clf.fit(y_true, y=y_true)
            best_params = None
            best_score = None
            search = None
            best_clf = clf

        model = best_clf.model_
        best_history = best_clf.history_

        if self.nn_method_ == "VAE":
            y_pred = model(
                tf.convert_to_tensor(y_train),
                training=False,
            )
            y_pred = self.tt_.inverse_transform(y_pred)
        elif self.nn_method_ == "SAE":
            y_pred = model(y_train, training=False)
            y_pred = self.tt_.inverse_transform(y_pred)
        elif self.nn_method_ in ["UBP", "NLPCA"]:
            # Third run_clf function
            y_pred_proba = model(model.V_latent, training=False)
            y_pred = self.tt_.inverse_transform(y_pred_proba)

        # Get metric scores.
        metrics = Scorers.scorer(
            y_true,
            y_pred,
            missing_mask=self.sim_missing_mask_,
            num_classes=self.num_classes,
            testing=self.testing,
        )

        if self.nn_method_ in ["UBP", "NLPCA"]:
            return (
                V,
                model,
                best_history,
                best_params,
                best_score,
                best_clf,
                search,
                metrics,
            )
        else:
            return (
                model,
                best_history,
                best_params,
                best_score,
                best_clf,
                search,
                metrics,
            )

    def _initialize_parameters(self, y_train):
        """Initialize important parameters.

        Args:
            y_train (numpy.ndarray): Training subset of original input data.

        Returns:
            Dict[str, Any]: Parameters to use for model.compile().
            Dict[str, Any]: Other parameters to pass to KerasClassifier().
            Dict[str, Any]: Parameters to pass to fit_params() in grid search.
        """
        # For CSVLogger() callback.

        append = True if self.nn_method_ == "UBP" else False
        logfile = os.path.join(
            f"{self.prefix}_output",
            "logs",
            "Unsupervised",
            self.nn_method_,
            "training_log.csv",
        )

        callbacks = [
            CSVLogger(filename=logfile, append=append),
            ReduceLROnPlateau(
                patience=self.lr_patience, min_lr=1e-6, min_delta=1e-6
            ),
        ]

        if self.nn_method_ in ["VAE", "SAE"]:
            callbacks.append(VAECallbacks())

            if self.nn_method_ == "VAE":
                callbacks.append(
                    CyclicalAnnealingCallback(
                        self.epochs, schedule_type="sigmoid"
                    )
                )
        else:
            callbacks.append(UBPCallbacks())

        search_mode = True if self.run_gridsearch_ else False

        if not self.disable_progressbar and not search_mode:
            callbacks.append(
                TqdmCallback(epochs=self.epochs, verbose=0, desc="Epoch: ")
            )

        if self.nn_method_ in ["UBP", "NLPCA"]:
            vinput = self._initV(y_train, search_mode)
            compile_params = self.nn_.set_compile_params(self.optimizer)
        else:
            vae = True if self.nn_method_ in ["VAE", "SAE"] else False

        if self.sample_weights == "auto" or self.sample_weights == "logsmooth":
            # Get class weights for each column.
            sample_weights = self.nn_.get_class_weights(
                self.y_original_,
                self.original_missing_mask_,
                return_1d=False,
                method=self.sample_weights,
            )
            sample_weights = self.nn_.normalize_data(sample_weights)

        elif isinstance(self.sample_weights, dict):
            for i in range(self.num_classes):
                if self.sample_weights[i] == 0.0:
                    self.sim_missing_mask_[self.y_original_ == i] = False

            sample_weights = self.nn_.get_class_weights(
                self.y_original_, user_weights=self.sample_weights
            )

        else:
            sample_weights = None

        vae = True if self.nn_method_ == "VAE" else False

        compile_params = self.nn_.set_compile_params(
            self.optimizer,
            sample_weights,
            vae=vae,
            act_func=self.act_func_,
        )

        compile_params["learning_rate"] = self.learning_rate

        if self.nn_method_ in ["VAE", "SAE"]:
            model_params = {
                "y": y_train,
                "batch_size": self.batch_size,
                "sample_weight": sample_weights,
                "missing_mask": self.original_missing_mask_,
                "output_shape": y_train.shape[1],
                "weights_initializer": self.weights_initializer,
                "n_components": self.n_components,
                "hidden_layer_sizes": self.hidden_layer_sizes,
                "num_hidden_layers": self.num_hidden_layers,
                "hidden_activation": self.hidden_activation,
                "l1_penalty": self.l1_penalty,
                "l2_penalty": self.l2_penalty,
                "dropout_rate": self.dropout_rate,
            }

            if self.nn_method_ == "VAE":
                model_params["kl_beta"] = (1.0 / y_train.shape[0],)
        else:
            model_params = {
                "V": vinput,
                "y_train": y_train,
                "batch_size": self.batch_size,
                "missing_mask": self.original_missing_mask_,
                "output_shape": y_train.shape[1],
                "weights_initializer": self.weights_initializer,
                "n_components": self.n_components,
                "hidden_layer_sizes": self.hidden_layer_sizes,
                "num_hidden_layers": self.num_hidden_layers,
                "hidden_activation": self.hidden_activation,
                "l1_penalty": self.l1_penalty,
                "l2_penalty": self.l2_penalty,
                "dropout_rate": self.dropout_rate,
                "num_classes": self.num_classes,
            }

        model_params["sample_weight"] = sample_weights

        fit_verbose = 1 if self.verbose == 2 else 0

        fit_params = {
            "batch_size": self.batch_size,
            "epochs": self.epochs,
            "callbacks": callbacks,
            "shuffle": True,
            "verbose": fit_verbose,
            "sample_weight": sample_weights,
        }

        if self.nn_method_ in ["VAE", "SAE"]:
            shuffle = True
            fit_params["validation_split"] = self.validation_split
        else:
            shuffle = False
            fit_params["validation_split"] = 0.0

        fit_params["shuffle"] = shuffle

        if self.run_gridsearch_ and "learning_rate" in self.gridparams:
            self.gridparams["optimizer__learning_rate"] = self.gridparams[
                "learning_rate"
            ]

            self.gridparams.pop("learning_rate")

        return (
            logfile,
            callbacks,
            compile_params,
            model_params,
            fit_params,
        )


class VAE(BaseNNImputer):
    """Class to impute missing data using a Variational Autoencoder neural network."""

    def __init__(
        self,
        kl_beta=tf.Variable(1.0, trainable=False),
        validation_split=0.2,
        **kwargs,
    ):
        self.kl_beta = kl_beta
        self.validation_split = validation_split

        self.nn_method_ = "VAE"
        self.num_classes = 4
        self.activate = None
        self.is_multiclass_ = True if self.num_classes != 4 else False
        self.testing = kwargs.get("testing", False)
        self.do_act_in_model_ = True if self.activate is None else False

        if self.do_act_in_model_ and self.is_multiclass_:
            self.act_func_ = "softmax"
        elif self.do_act_in_model_ and not self.is_multiclass_:
            self.act_func_ = "sigmoid"
        else:
            self.act_func_ = None

        super().__init__(
            self.activate,
            self.nn_method_,
            self.num_classes,
            self.act_func_,
            **kwargs,
            kl_beta=self.kl_beta,
            validation_split=self.validation_split,
        )

    def run_vae(
        self,
        y_true,
        y_train,
        model_params,
        compile_params,
        fit_params,
    ):
        """Run VAE using custom subclassed model.

        Args:
            y_true (numpy.ndarray): Original genotypes (training dataset) with known and missing values, of shape (n_samples, n_features).

            y_train (numpy.ndarray): Onehot encoded genotypes (training dataset) with known and missing values, of shape (n_samples, n_features, num_classes).

            model_params (Dict[str, Any]): Dictionary with parameters to pass to the classifier model.

            compile_params (Dict[str, Any]): Dictionary with parameters to pass to the tensorflow compile function.

            fit_params (Dict[str, Any]): Dictionary with parameters to pass to the fit() function.

        Returns:
            List[tf.keras.Model]: List of keras model objects. One for each phase (len=1 if NLPCA, len=3 if UBP).

            List[Dict[str, float]]: List of dictionaries with best neural network model history.

            Dict[str, Any] or None: Best parameters found during a grid search, or None if a grid search was not run.

            float: Best score obtained during grid search.

            tf.keras.Model: Best model found during grid search.

            sklearn.model_selection object (GridSearchCV, RandomizedSearchCV) or GASearchCV object.

            Dict[str, Any]: Per-class, micro, and macro-averaged metrics including accuracy, ROC-AUC, and Precision-Recall with Average Precision scores.
        """
        scorers = Scorers()
        scoring = None

        histories = list()
        models = list()

        if self.run_gridsearch_:
            scoring = scorers.make_multimetric_scorer(
                self.scoring_metrics_,
                self.sim_missing_mask_,
                num_classes=self.num_classes,
            )

        (
            model,
            best_history,
            best_params,
            best_score,
            best_clf,
            search,
            metrics,
        ) = self.run_clf(
            y_train,
            y_true,
            model_params,
            compile_params,
            fit_params,
            scoring=scoring,
        )

        histories.append(best_history)
        models.append(model)
        del model

        return (
            models,
            histories,
            best_params,
            best_score,
            best_clf,
            search,
            metrics,
        )


class SAE(BaseNNImputer):
    def __init__(
        self,
        **kwargs,
    ):
        self.num_classes = 3
        self.activate = "softmax"
        self.nn_method_ = "SAE"
        self.act_func_ = "softmax"
        self.testing = kwargs.get("testing", False)

        super().__init__(
            self.activate,
            self.nn_method_,
            self.num_classes,
            self.act_func_,
            **kwargs,
        )

    def run_sae(
        self,
        y_true,
        y_train,
        model_params,
        compile_params,
        fit_params,
    ):
        """Run standard autoencoder using custom subclassed model.

        Args:
            y_true (numpy.ndarray): Original genotypes (training dataset) with known and missing values of shape (n_samples, n_features).

            y_train (numpy.ndarray): Onehot-encoded genotypes (training dataset) with known and missing values of shape (n_samples, n_features, num_classes.)

            model_params (Dict[str, Any]): Dictionary with parameters to pass to the classifier model.

            compile_params (Dict[str, Any]): Dictionary with parameters to pass to the tensorflow compile function.

            fit_params (Dict[str, Any]): Dictionary with parameters to pass to the fit() function.

        Returns:
            List[tf.keras.Model]: List of keras model objects. One for each phase (len=1 if NLPCA, len=3 if UBP).

            List[Dict[str, float]]: List of dictionaries with best neural network model history.

            Dict[str, Any] or None: Best parameters found during a grid search, or None if a grid search was not run.

            float: Best score obtained during grid search.

            tf.keras.Model: Best model found during grid search.

            sklearn.model_selection object (GridSearchCV, RandomizedSearchCV) or GASearchCV object.

            Dict[str, Any]: Per-class, micro, and macro-averaged metrics including accuracy, ROC-AUC, and Precision-Recall with Average Precision scores.
        """
        scorers = Scorers()
        scoring = None

        histories = list()
        models = list()

        if self.run_gridsearch_:
            scoring = scorers.make_multimetric_scorer(
                self.scoring_metrics_, self.sim_missing_mask_
            )

        (
            model,
            best_history,
            best_params,
            best_score,
            best_clf,
            search,
            metrics,
        ) = self.run_clf(
            y_train,
            y_true,
            model_params,
            compile_params,
            fit_params,
            scoring=scoring,
            testing=False,
        )

        histories.append(best_history)
        models.append(model)
        del model

        return (
            models,
            histories,
            best_params,
            best_score,
            best_clf,
            search,
            metrics,
        )


class UBP(BaseNNImputer):
    def __init__(
        self,
        *,
        nlpca=False,
        **kwargs,
    ):
        # TODO: Make estimators compatible with variable number of classes.
        # E.g., with morphological data.
        self.nlpca = nlpca
        self.nn_method_ = "NLPCA" if self.nlpca else "UBP"
        self.num_classes = 3
        self.testing = kwargs.get("testing", False)
        self.activate = None
        self.act_func_ = "softmax"

        super().__init__(
            self.activate,
            self.nn_method_,
            self.num_classes,
            self.act_func_,
            **kwargs,
            nlpca=self.nlpca,
        )

    def run_nlpca(
        self,
        y_true,
        y_train,
        model_params,
        compile_params,
        fit_params,
    ):
        """Run NLPCA using custom subclassed model.

        Args:
            y_true (numpy.ndarray): Original genotypes with known and missing values.

            y_train (numpy.ndarray): For compatibility with VAE and SAE. Not used.

            model_params (Dict[str, Any]): Dictionary with parameters to pass to the classifier model.

            compile_params (Dict[str, Any]): Dictionary with parameters to pass to the tensorflow compile function.

            fit_params (Dict[str, Any]): Dictionary with parameters to pass to the fit() function.

        Returns:
            List[tf.keras.Model]: List of keras model objects. One for each phase (len=1 if NLPCA, len=3 if UBP).

            List[Dict[str, float]]: List of dictionaries with best neural network model history.

            Dict[str, Any] or None: Best parameters found during a grid search, or None if a grid search was not run.

            float: Best score obtained during grid search.

            tf.keras.Model: Best model found during grid search.

            sklearn.model_selection object (GridSearchCV, RandomizedSearchCV) or GASearchCV object.

            Dict[str, Any]: Per-class, micro, and macro-averaged metrics including accuracy, ROC-AUC, and Precision-Recall with Average Precision scores.
        """
        scorers = Scorers()

        histories = list()
        models = list()
        y_train = model_params.pop("y_train")
        ubp_weights = None
        phase = None
        scoring = None

        if self.run_gridsearch_:
            scoring = scorers.make_multimetric_scorer(
                self.scoring_metrics_, self.sim_missing_mask_
            )

        (
            V,
            model,
            best_history,
            best_params,
            best_score,
            best_clf,
            search,
            metrics,
        ) = self.run_clf(
            y_train,
            y_true,
            model_params,
            compile_params,
            fit_params,
            ubp_weights=ubp_weights,
            phase=phase,
            scoring=scoring,
            testing=False,
        )

        histories.append(best_history)
        models.append(model)
        del model

        return (
            models,
            histories,
            best_params,
            best_score,
            best_clf,
            search,
            metrics,
        )

    def run_ubp(
        self,
        y_true,
        y_train,
        model_params,
        compile_params,
        fit_params,
    ):
        """Run UBP using custom subclassed model.

        Args:
            y_true (numpy.ndarray): Original genotypes with known and missing values.

            y_train (numpy.ndarray): For compatibility with VAE and SAE. Not used.

            model_params (Dict[str, Any]): Dictionary with parameters to pass to the classifier model.

            compile_params (Dict[str, Any]): Dictionary with parameters to pass to the tensorflow compile function.

            fit_params (Dict[str, Any]): Dictionary with parameters to pass to the fit() function.

        Returns:
            List[tf.keras.Model]: List of keras model objects. One for each phase (len=1 if NLPCA, len=3 if UBP).

            List[Dict[str, float]]: List of dictionaries with best neural network model history.

            Dict[str, Any] or None: Best parameters found during a grid search, or None if a grid search was not run.

            float: Best score obtained during grid search.

            tf.keras.Model: Best model found during grid search.

            sklearn.model_selection object (GridSearchCV, RandomizedSearchCV) or GASearchCV object.

            Dict[str, Any]: Per-class, micro, and macro-averaged metrics including accuracy, ROC-AUC, and Precision-Recall with Average Precision scores.
        """
        scorers = Scorers()

        histories = list()
        models = list()
        search_n_components = False

        y_train = model_params.pop("y_train")

        if self.run_gridsearch_:
            # Cannot do CV because there is no way to use test splits
            # given that the input gets refined. If using a test split,
            # then it would just be the randomly initialized values and
            # would not accurately represent the model.
            # Thus, we disable cross-validation for the grid searches.
            scoring = scorers.make_multimetric_scorer(
                self.scoring_metrics_, self.sim_missing_mask_
            )

            if "n_components" in self.gridparams:
                search_n_components = True
                n_components_searched = self.n_components
        else:
            scoring = None

        for phase in range(1, 4):
            ubp_weights = models[1].get_weights() if phase == 3 else None

            (
                V,
                model,
                best_history,
                best_params,
                best_score,
                best_clf,
                search,
                metrics,
            ) = self.run_clf(
                y_train,
                y_true,
                model_params,
                compile_params,
                fit_params,
                ubp_weights=ubp_weights,
                phase=phase,
                scoring=scoring,
                testing=False,
            )

            if phase == 1:
                # Cannot have V input with different n_components
                # in other phases than are in phase 1.
                # So the n_components search has to happen in phase 1.
                if best_params is not None and search_n_components:
                    n_components_searched = best_params["n_components"]
                    model_params["V"] = {
                        n_components_searched: model.V_latent.copy()
                    }
                    model_params["n_components"] = n_components_searched
                    self.n_components = n_components_searched
                    self.gridparams.pop("n_components")

                else:
                    model_params["V"] = V
            elif phase == 2:
                model_params["V"] = V

            elif phase == 3:
                if best_params is not None and search_n_components:
                    best_params["n_components"] = n_components_searched

            histories.append(best_history)
            models.append(model)
            del model

        return (
            models,
            histories,
            best_params,
            best_score,
            best_clf,
            search,
            metrics,
        )

    def _initV(self, y_train, search_mode):
        """Initialize random input V as dictionary of numpy arrays.

        Args:
            y_train (numpy.ndarray): One-hot encoded training dataset (actual data).

            search_mode (bool): Whether doing grid search.

        Returns:
            Dict[int, numpy.ndarray]: Dictionary with n_components: V as key-value pairs.

        Raises:
            ValueError: Number of components must be >= 2.
        """
        vinput = dict()
        if search_mode:
            if "n_components" in self.gridparams:
                n_components = self.gridparams["n_components"]
            else:
                n_components = self.n_components

            if not isinstance(n_components, int):
                if min(n_components) < 2:
                    raise ValueError(
                        f"n_components must be >= 2, but a value of {n_components} was specified."
                    )

                elif len(n_components) == 1:
                    vinput[n_components[0]] = self.nn_.init_weights(
                        y_train.shape[0], n_components[0]
                    )

                else:
                    for c in n_components:
                        vinput[c] = self.nn_.init_weights(y_train.shape[0], c)
            else:
                vinput[self.n_components] = self.nn_.init_weights(
                    y_train.shape[0], self.n_components
                )

        else:
            vinput[self.n_components] = self.nn_.init_weights(
                y_train.shape[0], self.n_components
            )

        return vinput



import sys
import os
import copy
import unittest
import pprint
from pgsui.impute.estimators import (
    ImputeKNN,
    ImputeRandomForest,
    ImputeXGBoost,
    ImputeVAE,
    ImputeStandardAutoEncoder,
    ImputeUBP,
    ImputeNLPCA,
)
from pgsui.impute.simple_imputers import (
    ImputePhylo,
    ImputeMF,
    ImputeAlleleFreq,
    ImputeRefAllele,
)

from snpio import GenotypeData
from pgsui.data_processing.transformers import SimGenotypeDataTransformer
import numpy as np


class HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


class TestMyClasses(unittest.TestCase):
    def setUp(self):
        with HiddenPrints():
            self.genotype_data = GenotypeData(
                filename="pgsui/example_data/phylip_files/test_n100.phy",
                popmapfile="pgsui/example_data/popmaps/test.popmap",
                guidetree="pgsui/example_data/trees/test.tre",
                qmatrix="pgsui/example_data/trees/test.qmat",
                siterates="pgsui/example_data/trees/test_siterates_n100.txt",
                prefix="test_imputer",
                force_popmap=True,
                plot_format="png",
            )

            # Create a SimGenotypeDataTransformer instance and use it
            # to simulate missing data
            self.transformer = SimGenotypeDataTransformer(
                genotype_data=self.genotype_data,
                prop_missing=0.2,
                strategy="random_weighted",
            )
            self.transformer.fit(self.genotype_data.genotypes_012(fmt="numpy"))
            self.simulated_data = copy.deepcopy(self.genotype_data)

            self.simulated_data.genotypes_012 = self.transformer.transform(
                self.genotype_data.genotypes_012(fmt="numpy")
            )

    def _test_class(self, class_instance, do_gridsearch=False):
        print(f"\nMETHOD: {class_instance.__name__}\n")

        if do_gridsearch:
            # Do a simple test.
            if class_instance in [ImputeRandomForest, ImputeXGBoost]:
                param_grid = {"n_estimators": [50, 100]}  # Do a simple test
            elif class_instance in [
                ImputeVAE,
                ImputeStandardAutoEncoder,
                ImputeNLPCA,
                ImputeUBP,
            ]:
                param_grid = {"dropout_rate": [0.1, 0.2]}
            elif class_instance == ImputeKNN:
                param_grid = {"n_neighbors": [5, 8]}
        else:
            param_grid = None

        kwargs = {
            'sim_strategy': "random_weighted"
        }

        instance = class_instance(
            self.simulated_data, 
            gridparams=param_grid, 
            **kwargs)

        imputed_data = instance.imputed.genotypes_012(fmt="numpy")

        # Test that the imputed values are close to the original values
        accuracy = self.transformer.accuracy(
            self.genotype_data.genotypes_012(fmt="numpy"), imputed_data
        )

        (
            auc_roc_scores,
            precision_scores,
            recall_scores,
            avg_precision_scores,
        ) = self.transformer.auc_roc_pr_ap(
            self.genotype_data.genotypes_012(fmt="numpy"), imputed_data
        )

        pprint.PrettyPrinter(indent=4, sort_dicts=True).pprint(
            f"OVERALL ACCURACY: {accuracy}"
        )
        pprint.PrettyPrinter(indent=4, sort_dicts=True).pprint(
            f"AUC-ROC PER CLASS: {dict(zip(range(3), auc_roc_scores))}"
        )
        pprint.PrettyPrinter(indent=4, sort_dicts=True).pprint(
            f"PRECISION PER CLASS: {dict(zip(range(3), precision_scores))}"
        )
        pprint.PrettyPrinter(indent=4, sort_dicts=True).pprint(
            f"RECALL PER CLASS: {dict(zip(range(3), recall_scores))}"
        )
        pprint.PrettyPrinter(indent=4, sort_dicts=True).pprint(
            f"AVERAGE PRECISION PER CLASS: {dict(zip(range(3), avg_precision_scores))}"
        )
        print("\n")

    # def test_ImputeKNN(self):
    #     self._test_class(ImputeKNN)

    # def test_ImputeRandomForest(self):
    #     self._test_class(ImputeRandomForest)

    # def test_ImputeXGBoost(self):
    #     self._test_class(ImputeXGBoost)

    # def test_ImputeVAE(self):
    #     self._test_class(ImputeVAE)

    # def test_ImputeStandardAutoEncoder(self):
    #     self._test_class(ImputeStandardAutoEncoder)

    def test_ImputeUBP(self):
        self._test_class(ImputeUBP)

    # def test_ImputeNLPCA(self):
    #     self._test_class(ImputeNLPCA)

    # def test_ImputeKNN_grid(self):
    #     self._test_class(ImputeKNN, do_gridsearch=True)

    # def test_ImputeRandomForest_grid(self):
    #     self._test_class(ImputeRandomForest, do_gridsearch=True)

    # def test_ImputeXGBoost_grid(self):
    #     self._test_class(ImputeXGBoost, do_gridsearch=True)

    # def test_ImputeVAE_grid(self):
    #     self._test_class(ImputeVAE, do_gridsearch=True)

    # def test_ImputeStandardAutoEncoder_grid(self):
    #     self._test_class(ImputeStandardAutoEncoder, do_gridsearch=True)

    # def test_ImputeUBP_grid(self):
    #     self._test_class(ImputeUBP, do_gridsearch=True)

    # def test_ImputeNLPCA_grid(self):
    #     self._test_class(ImputeNLPCA, do_gridsearch=True)

    # def test_ImputePhylo(self):
    #     self._test_class(ImputePhylo)

    # def test_ImputeAlleleFreq(self):
    #     self._test_class(ImputeAlleleFreq)

    # def test_ImputeRefAllele(self):
    #     self._test_class(ImputeRefAllele)

    # def test_ImputeMF(self):
    #     self._test_class(ImputeMF)

    # def test_ImputeRefAllele(self):
    #     self._test_class(ImputeRefAllele)


if __name__ == "__main__":
    unittest.main()



import sys
import os
import functools
import time
import datetime
import platform
import subprocess
import re
import logging

import numpy as np
import pandas as pd
from tqdm import tqdm
from tqdm.utils import disp_len, _unicode  # for overriding status_print


# from skopt import BayesSearchCV


def validate_input_type(X, return_type="array"):
    """Validate input type and return as numpy array.

    Args:
        X (pandas.DataFrame, numpy.ndarray, or List[List[int]]): Input data.

        return_type (str): Type of returned object. Supported options include: "df", "array", and "list". Defaults to "array".

    Returns:
        pandas.DataFrame, numpy.ndarray, or List[List[int]]: Input data desired return_type.

    Raises:
        TypeError: X must be of type pandas.DataFrame, numpy.ndarray, or List[List[int]].

        ValueError: Unsupported return_type provided. Supported types are "df", "array", and "list".

    """
    if not isinstance(X, (pd.DataFrame, np.ndarray, list)):
        raise TypeError(
            f"X must be of type pandas.DataFrame, numpy.ndarray, "
            f"or List[List[int]], but got {type(X)}"
        )

    if return_type == "array":
        if isinstance(X, pd.DataFrame):
            return X.to_numpy()
        elif isinstance(X, list):
            return np.array(X)
        elif isinstance(X, np.ndarray):
            return X.copy()

    elif return_type == "df":
        if isinstance(X, pd.DataFrame):
            return X.copy()
        elif isinstance(X, (np.ndarray, list)):
            return pd.DataFrame(X)

    elif return_type == "list":
        if isinstance(X, list):
            return X
        elif isinstance(X, np.ndarray):
            return X.tolist()
        elif isinstance(X, pd.DataFrame):
            return X.values.tolist()

    else:
        raise ValueError(
            f"Unsupported return type provided: {return_type}. Supported types "
            f"are 'df', 'array', and 'list'"
        )


def generate_random_dataset(
    min_value=0,
    max_value=2,
    nrows=35,
    ncols=20,
    min_missing_rate=0.15,
    max_missing_rate=0.5,
):
    """Generate a random integer dataset that can be used for testing.

    Will also add randomly missing values of random proportions between ``min_missing_rate`` and ``max_missing_rate``.

    Args:
        min_value (int, optional): Minimum value to use. Defaults to 0.

        max_value (int, optional): Maxiumum value to use. Defaults to 2.

        nrows (int, optional): Number of rows to use. Defaults to 35.

        ncols (int, optional): Number of columns to use. Defaults to 20.

        min_missing_rate (float, optional): Minimum proportion of missing data per column. Defaults to 0.15.

        max_missing_rate (float, optional): Maximum proportion of missing data per column.

    Returns:
        numpy.ndarray: Numpy array with randomly generated dataset.
    """
    assert (
        min_missing_rate >= 0 and min_missing_rate < 1.0
    ), f"min_missing_rate must be >= 0 and < 1.0, but got {min_missing_rate}"

    assert (
        max_missing_rate > 0 and max_missing_rate < 1.0
    ), f"max_missing_rate must be > 0 and < 1.0, but got {max_missing_rate}"

    assert nrows > 1, f"nrows must be > 1, but got {nrows}"
    assert ncols > 1, f"ncols must be > 1, but got {ncols}"

    try:
        min_missing_rate = float(min_missing_rate)
        max_missing_rate = float(max_missing_rate)
    except TypeError:
        sys.exit(
            "min_missing_rate and max_missing_rate must be of type float or "
            "must be cast-able to type float"
        )

    X = np.random.randint(
        min_value, max_value + 1, size=(nrows, ncols)
    ).astype(float)
    for i in range(X.shape[1]):
        drop_rate = int(
            np.random.choice(
                np.arange(min_missing_rate, max_missing_rate, 0.02), 1
            )[0]
            * X.shape[0]
        )

        rows = np.random.choice(np.arange(0, X.shape[0]), size=drop_rate)
        X[rows, i] = np.nan

    return X


def generate_012_genotypes(
    nrows=35,
    ncols=20,
    max_missing_rate=0.5,
    min_het_rate=0.001,
    max_het_rate=0.3,
    min_alt_rate=0.001,
    max_alt_rate=0.3,
):
    """Generate random 012-encoded genotypes.

    Allows users to control the rate of reference, heterozygote, and alternate alleles. Will insert a random proportion between ``min_het_rate`` and ``max_het_rate`` and ``min_alt_rate`` and ``max_alt_rate`` and from no misssing data to a proportion of ``max_missing_rate``.

    Args:
        nrows (int, optional): Number of rows to generate. Defaults to 35.

        ncols (int, optional): Number of columns to generate. Defaults to 20.

        max_missing_rate (float, optional): Maximum proportion of missing data to use. Defaults to 0.5.

        min_het_rate (float, optional): Minimum proportion of heterozygotes (1's) to insert. Defaults to 0.001.

        max_het_rate (float, optional): Maximum proportion of heterozygotes (1's) to insert. Defaults to 0.3.

        min_alt_rate (float, optional): Minimum proportion of alternate alleles (2's) to insert. Defaults to 0.001.

        max_alt_rate (float, optional): Maximum proportion of alternate alleles (2's) to insert. Defaults to 0.3.
    """
    assert (
        min_het_rate > 0 and min_het_rate <= 1.0
    ), f"min_het_rate must be > 0 and <= 1.0, but got {min_het_rate}"

    assert (
        max_het_rate > 0 and max_het_rate <= 1.0
    ), f"max_het_rate must be > 0 and <= 1.0, but got {max_het_rate}"

    assert (
        min_alt_rate > 0 and min_alt_rate <= 1.0
    ), f"min_alt_rate must be > 0 and <= 1.0, but got {min_alt_rate}"

    assert (
        max_alt_rate > 0 and max_alt_rate <= 1.0
    ), f"max_alt_rate must be > 0 and <= 1.0, but got {max_alt_rate}"

    assert nrows > 1, f"The number of rows must be > 1, but got {nrows}"

    assert ncols > 1, f"The number of columns must be > 1, but got {ncols}"

    assert (
        max_missing_rate > 0 and max_missing_rate < 1.0
    ), f"max_missing rate must be > 0 and < 1.0, but got {max_missing_rate}"

    try:
        min_het_rate = float(min_het_rate)
        max_het_rate = float(max_het_rate)
        min_alt_rate = float(min_alt_rate)
        max_alt_rate = float(max_alt_rate)
        max_missing_rate = float(max_missing_rate)
    except TypeError:
        sys.exit(
            "max_missing_rate, min_het_rate, max_het_rate, min_alt_rate, and "
            "max_alt_rate must be of type float, or must be cast-able to type "
            "float"
        )

    X = np.zeros((nrows, ncols))
    for i in range(X.shape[1]):
        het_rate = int(
            np.ceil(
                np.random.choice(
                    np.arange(min_het_rate, max_het_rate, 0.02), 1
                )[0]
                * X.shape[0]
            )
        )

        alt_rate = int(
            np.ceil(
                np.random.choice(
                    np.arange(min_alt_rate, max_alt_rate, 0.02), 1
                )[0]
                * X.shape[0]
            )
        )

        het = np.sort(
            np.random.choice(
                np.arange(0, X.shape[0]), size=het_rate, replace=False
            )
        )

        alt = np.sort(
            np.random.choice(
                np.arange(0, X.shape[0]), size=alt_rate, replace=False
            )
        )

        sidx = alt.argsort()
        idx = np.searchsorted(alt, het, sorter=sidx)
        idx[idx == len(alt)] = 0
        het_unique = het[alt[sidx[idx]] != het]

        X[alt, i] = 2
        X[het_unique, i] = 1

        drop_rate = int(
            np.random.choice(np.arange(0.15, max_missing_rate, 0.02), 1)[0]
            * X.shape[0]
        )

        missing = np.random.choice(np.arange(0, X.shape[0]), size=drop_rate)

        X[missing, i] = np.nan

    print(
        f"Created a dataset of shape {X.shape} with {np.isnan(X).sum()} total missing values"
    )

    return X


def unique2D_subarray(a):
    """Get unique subarrays for each column from numpy array."""
    dtype1 = np.dtype((np.void, a.dtype.itemsize * np.prod(a.shape[1:])))
    b = np.ascontiguousarray(a.reshape(a.shape[0], -1)).view(dtype1)
    return a[np.unique(b, return_index=1, axis=-1)[1]]


def _remove_nonbiallelic(df, cv=5):
    """Remove non-biallelic sites from pandas.DataFrame.

    Remove sites that do not have both 0 and 2 encoded values in a column and if any of the allele counts is less than the number of cross-validation folds.

    Args:
        df (pandas.DataFrame): DataFrame with 012-encoded genotypes.

    Returns:
        pandas.DataFrame: DataFrame with non-biallelic sites dropped.
    """
    df_cp = df.copy()
    bad_cols = list()
    if pd.__version__[0] == 0:
        for col in df_cp.columns:
            if (
                not df_cp[col].isin([0.0]).any()
                or not df_cp[col].isin([2.0]).any()
            ):
                bad_cols.append(col)

            elif len(df_cp[df_cp[col] == 0.0]) < cv:
                bad_cols.append(col)

            elif df_cp[col].isin([1.0]).any():
                if len(df_cp[df_cp[col] == 1]) < cv:
                    bad_cols.append(col)

            elif len(df_cp[df_cp[col] == 2.0]) < cv:
                bad_cols.append(col)

    # pandas 1.X.X
    else:
        for col in df_cp.columns:
            if 0.0 not in df[col].unique() and 2.0 not in df[col].unique():
                bad_cols.append(col)

            elif len(df_cp[df_cp[col] == 0.0]) < cv:
                bad_cols.append(col)

            elif 1.0 in df_cp[col].unique():
                if len(df_cp[df_cp[col] == 1.0]) < cv:
                    bad_cols.append(col)

            elif len(df_cp[df_cp[col] == 2.0]) < cv:
                bad_cols.append(col)

    if bad_cols:
        df_cp.drop(bad_cols, axis=1, inplace=True)

        print(
            f"{len(bad_cols)} columns removed for being non-biallelic or "
            f"having genotype counts < number of cross-validation "
            f"folds\nSubsetting from {len(df_cp.columns)} remaining columns\n"
        )

    return df_cp


def get_indices(l):
    """Takes a list and returns dict giving indices matching each possible
    list member.

    Example:
            Input [0, 1, 1, 0, 0]
            Output {0:[0,3,4], 1:[1,2]}
    """
    ret = dict()
    for member in set(l):
        ret[member] = list()
    i = 0
    for el in l:
        ret[el].append(i)
        i += 1
    return ret


def all_zero(l):
    """Check whether list consists of all zeros.

    Returns TRUE if supplied list contains all zeros
    Returns FALSE if list contains ANY non-zero values
    Returns FALSE if list is empty.

    Args:
        l (List[int]): List to check.

    Returns:
        bool: True if all zeros, False otherwise.
    """
    values = set(l)
    if len(values) > 1:
        return False
    elif len(values) == 1 and l[0] in [0, 0.0, "0", "0.0"]:
        return True
    else:
        return False


def weighted_draw(d, num_samples=1):
    choices = list(d.keys())
    weights = list(d.values())
    return np.random.choice(choices, num_samples, p=weights)


def timer(func):
    """print the runtime of the decorated function in the format HH:MM:SS."""

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()
        value = func(*args, **kwargs)
        end_time = time.perf_counter()
        run_time = end_time - start_time
        final_runtime = str(datetime.timedelta(seconds=run_time))
        print(f"Finshed {func.__name__!r} in {final_runtime}\n")
        return value

    return wrapper_timer


def progressbar(it, prefix="", size=60, file=sys.stdout):
    count = len(it)

    def show(j):
        x = int(size * j / count)
        file.write(
            "%s[%s%s] %i/%i\r" % (prefix, "#" * x, "." * (size - x), j, count)
        )
        file.flush()

    show(0)
    for i, item in enumerate(it):
        yield item
        show(i + 1)
    file.write("\n")
    file.flush()


def isnotebook():
    """Checks whether in Jupyter notebook.

    Returns:
        bool: True if in Jupyter notebook, False otherwise.
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            # Jupyter notebook or qtconsole
            return True
        elif shell == "TerminalInteractiveShell":
            # Terminal running IPython
            return False
        else:
            # Other type (?)
            return False
    except NameError:
        # Probably standard Python interpreter
        return False


def get_processor_name():
    if platform.system() == "Windows":
        return platform.processor()
    elif platform.system() == "Darwin":
        # os.environ["PATH"] = os.environ["PATH"] + os.pathsep + "/usr/sbin"
        arch = platform.processor()
        if arch[0] == "i":
            return "Intel"
        else:
            return arch
    elif platform.system() == "Linux":
        command = "cat /proc/cpuinfo"
        all_info = subprocess.check_output(command, shell=True).strip()
        all_info = all_info.decode("utf-8")
        for line in all_info.split("\n"):
            if "model name" in line:
                return re.sub(".*model name.*:", "", line, 1)
    return ""


class tqdm_linux(tqdm):
    """Adds a dynamically updating progress bar.

    Decorate an iterable object, with a dynamically updating progressbar every time a value is requested.
    """

    @staticmethod
    def status_printer(self, file):
        """Manage the printing and in-place updating of a line of characters.

        NOTE: If the string is longer than a line, then in-place updating may not work (it will print a new line at each refresh).

        Overridden to work with linux HPC clusters. Replaced carriage return with linux newline in fp_write function.

        Args:
            file (str): Path of file to print status to.
        """

        fp = file
        fp_flush = getattr(fp, "flush", lambda: None)

        def fp_write(s):
            fp.write(_unicode(s))
            fp_flush()

        last_len = [0]

        def print_status(s):
            len_s = disp_len(s)
            fp_write("\n" + s + (" " * max(last_len[0] - len_s, 0)))
            last_len[0] = len_s

        return print_status


class HiddenPrints:
    """Class to supress printing within a with statement."""

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


class StreamToLogger(object):
    """Fake file-like stream object that redirects writes to a logger instance."""

    def __init__(self, logger, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ""

    def write(self, buf):
        temp_linebuf = self.linebuf + buf
        self.linebuf = ""
        for line in temp_linebuf.splitlines(True):
            # From the io.TextIOWrapper docs:
            #   On output, if newline is None, any '\n' characters written
            #   are translated to the system default line separator.
            # By default sys.stdout.write() expects '\n' newlines and then
            # translates them so this is still cross platform.
            if line[-1] == "\n":
                self.logger.log(self.log_level, line.rstrip())
            else:
                self.linebuf += line

    def flush(self):
        if self.linebuf != "":
            self.logger.log(self.log_level, self.linebuf.rstrip())
        self.linebuf = ""



#!/usr/bin/env python
import sys
import os
import subprocess
import errno

import toytree
import toyplot.pdf
import pyvolve
import copy
import random
import re
import numpy as np
import pandas as pd

import matplotlib.pyplot as plt

def main():
    """
    Using pyvolve and toytree to simulate data for PG-SUI

    Generates random data matrices w/ varying dimensions
    for the purposes of runtime benchmarking

    """
    seed=553423
    random.seed(seed)

    num_samples_range=[10, 100, 1000]
    num_loci_range=[100, 1000, 10000]
    loc_length=100
    write_gene_alignments=False
    make_gene_trees=False
    make_guidetrees=True #set to true to run IQTREE on simulated SNP matrices
    keep_all=False #set to true to keep ALL iqtree outputs
    keep_report=True #set to true to keep .iqtree files
    get_siterates=True #set to true to infer site-specific rates in IQTREE
    snps_per_locus=1
    iqtree_bin="iqtree2"
    get_rates=True
    iq_procs=2

    ###################

    if get_siterates and not make_guidetrees:
        print("ERROR: can't set get_siterates=True and make_guidetrees=False")
        print("Setting make_guidetrees=True and proceeding...")
        make_guidetrees=True

    for num_samples in num_samples_range:
        outgroup="r0"
        tobj = toytree.rtree.unittree(ntips=num_samples,
                                    treeheight=1.0,
                                    random_names=False,
                                    seed=random.randint(1, (sys.maxsize * 2 + 1)))
        guidetree=tobj.write(tree_format=5)
        for num_loci in num_loci_range:
            print("num_samples:",num_samples)
            print("num_loci:",num_loci)
            pass

            base="benchmark_"+"i"+str(num_samples)+"_l"+str(num_loci)
            tobj.write((base+"_guidetree.tre"))

            data=dict()
            for ind in tobj.get_tip_labels():
                data[ind] = list()

            my_tree = pyvolve.read_tree(tree=guidetree)


            #for model in ["gtr","gtrgamma"]:
            for model in ["gtrgamma"]:
                model_outpath=base+"_"+model

                for locus in range(num_loci):
                    print(locus)
                    f = np.random.random(4)
                    f /= f.sum()
                    parameters = {
                    "mu":
                        {"AC": np.random.uniform(low=0.0, high=1.0),
                        "AG": np.random.uniform(low=0.0, high=1.0),
                        "AT": np.random.uniform(low=0.0, high=1.0),
                        "CG": np.random.uniform(low=0.0, high=1.0),
                        "CT": np.random.uniform(low=0.0, high=1.0),
                        "GT": np.random.uniform(low=0.0, high=1.0)},
                    "state_freqs":
                        [f[0], f[1], f[2], f[3]]
                    }
                    if model == "gtr":
                        #GTR model, without rate heterogeneity
                        my_model = pyvolve.Model("nucleotide",
                            parameters)
                    else:
                        my_model = pyvolve.Model("nucleotide",
                            parameters,
                            rate_factors = [
                                        np.random.uniform(low=0.1, high=0.7, size=1),
                                        np.random.uniform(low=0.5, high=1.2, size=1),
                                        np.random.uniform(low=1.0, high=1.8, size=1),
                                        np.random.uniform(low=1.5, high=5.0, size=1)
                                        ],
                            rate_probs = [0.4, 0.3, 0.2, 0.1] )
                    if write_gene_alignments:
                        fasta_outpath="full_alignments/"
                        if not os.path.exists(fasta_outpath):
                            os.mkdir(fasta_outpath)
                    else:
                        fasta_outpath=model_outpath
                    fastaout=fasta_outpath +model_outpath+"_loc"+str(locus) + "_gene-alignment.fasta"
                    #sample a gene alignment
                    loc = sample_locus(my_tree, my_model, loc_length, snps_per_locus, fastaout)

                    if loc:
                        #sample SNP(s) from gene alignment
                        sampled = sample_snp(read_fasta(fastaout), loc_length, snps_per_locus)
                        if sampled is not None:
                            data = add_locus(data,sampled)

                        if not write_gene_alignments:
                            os.remove(fastaout)
                            if make_gene_trees:
                                print("ERROR: Can't make gene trees when write_gene_alignments = False")
                        elif make_gene_trees:
                            run_iqtree(fastaout,
                                iqtree_path=iqtree_bin,
                                keep_all=keep_all,
                                keep_report=keep_report,
                                rates=get_siterates,
                                procs=iq_procs)
                            reroot_tree(tree=(fastaout+".treefile"),
                                rooted=(fastaout+".rooted.tre"),
                                outgroup_wildcard=outgroup)

                #write full SNP alignment & generate tree
                all_snp_out=model_outpath+"_base-snps-concat.fasta"
                write_fasta(data, all_snp_out)
                if make_guidetrees:
                    run_iqtree(all_snp_out,
                        iqtree_path=iqtree_bin,
                        keep_all=keep_all,
                        keep_report=keep_report,
                        rates=get_siterates,
                        procs=iq_procs)
                    reroot_tree(tree=(all_snp_out+".treefile"),
                        rooted=(all_snp_out+".rooted.tre"),
                        outgroup_wildcard=outgroup)

def reroot_tree(tree, rooted="out.rooted.tre", outgroup_wildcard="out"):
    t=toytree.tree(tree)
    try:
        rt=t.root(wildcard=outgroup_wildcard)
        rt.write(rooted, tree_format=5)
        return(rt)
    except Exception:
        t.write(rooted, tree_format=5)
        return(None)


def hybridization(dat, prob=0.1, source=None, target=None):
    new_dat=dict()
    if source is None:
        source = [key for key in dat.keys()]
    if target is None:
        target = [key for key in dat.keys()]

    for individual in dat.keys():
        new_dat[individual] = dat[individual]
        aln_len=len(dat[individual])
    all_indices=list(range(aln_len))
    num=int(aln_len*prob)

    for target_individual in target:
        snp_indices = np.random.choice(all_indices, size=num, replace=False)
        for index in snp_indices:
            source_ind=np.random.choice(source, size=1)[0]
            new_dat[target_individual][index] = new_dat[source_ind][index]
    return(new_dat)

def run_iqtree(aln,
    iqtree_path="iqtree",
    keep_all=False,
    keep_report=False,
    outgroup=None,
    rates=False,
    procs=4):
    #run
    cmd = [iqtree_path,
                    "-s",
                    str(aln),
                    "-m",
                    "GTR+I*G4",
                    "-redo",
                    "-T",
                    str(procs)
                    ]
    if outgroup is not None:
        cmd.append("-o")
        cmd.append(str(outgroup))
    if rates:
        #use -wst (NOT -mlrate) if using iq-tree 1.6xx
        #cmd.append("-wsr")
        cmd.append("--mlrate")
    result = subprocess.run(cmd, capture_output=True, text=True)
    #print(result.stdout)
    #print(result.stderr)

    if not keep_all:
        #delete everything except treefile
        silentremove((aln + ".bionj"))
        silentremove((aln + ".ckp.gz"))
        silentremove((aln + ".log"))
        silentremove((aln + ".mldist"))
        silentremove((aln + ".uniqueseq.phy"))
        if not keep_report:
            silentremove((aln + ".iqtree"))
    return((aln + ".treefile"))

def add_locus(d, new):
    for sample in d.keys():
        for snp in new[sample]:
            d[sample].append(snp)
    return(d)

def write_fasta(seqs, fas):
	with open(fas, 'w') as fh:
		#Write seqs to FASTA first
		for a in seqs.keys():
			name = ">" + str(a) + "\n"
			seq = "".join(seqs[a]) + "\n"
			fh.write(name)
			fh.write(seq)
		fh.close()

def read_fasta(fasta):
    data = dict()
    header=False
    sample=None
    sequence=""
    with open(fasta, "r") as fin:
        for line in fin:
            line = line.strip()
            if not line:  # If blank line.
                continue
            if line[0] == ">":
                if sample:
                    data[sample] = sequence
                    sequence = ""
                sample=line[1:]
            else:
                sequence = sequence + line
        data[sample] = sequence
        return(data)

def sample_snp(aln_dict, aln_len, snps_per_locus=1):
    snp_indices = []
    snp_aln = dict()
    if aln_len == 1:
        for sample in aln_dict.keys():
            snp_aln[sample] = aln_dict[sample][0]
        return(snp_aln)
    else:
        for sample in aln_dict.keys():
            snp_aln[sample] = []
    for i in range(aln_len):
        vars=[]
        for sample in aln_dict.keys():
            nuc=aln_dict[sample][i]
            if len(vars) == 0:
                vars.append(nuc)
            elif nuc not in vars:
                snp_indices.append(i)
                break
    if len(snp_indices) == 0:
        return(None)
    elif len(snp_indices) == 1:
        #sample them all
        for sample in aln_dict.keys():
            snp_aln[sample] = aln_dict[sample][snp_indices[0]]
    else:
        sampled_indices = np.random.choice(snp_indices, size=snps_per_locus, replace=False)
        for sample in aln_dict.keys():
            for i in sampled_indices:
                snp_aln[sample].append(aln_dict[sample][i])
    return(snp_aln)

def sample_locus(tree, model, gene_len=1000, num_snps=1, out="out.fasta"):
    try:
        my_partition = pyvolve.Partition(models = model, size=gene_len)
        my_evolver = pyvolve.Evolver(partitions = my_partition, tree = tree)
        my_evolver(seqfile = out,
            seqfmt = "fasta",
            ratefile=False,
            infofile=False)
        return(True)
    except Exception:
        return False

def silentremove(filename):
    try:
        os.remove(filename)
    except OSError as e: # this would be "except OSError, e:" before Python 2.6
        if e.errno != errno.ENOENT: # errno.ENOENT = no such file or directory
            raise # re-raise exception if a different error occurred

def get_tree_tips(tree):
    tips = re.split('[ ,\(\);]', tree)
    return([i for i in tips if i])

def basic_tree_plot(tree, out="out.pdf"):
    mystyle = {
        "edge_type": "p",
        "edge_style": {
            "stroke": toytree.colors[0],
            "stroke-width": 1,
        },
        "tip_labels_align": True,
        "tip_labels_style": {"font-size": "5px"},
        "node_labels": False,
        "tip_labels": True
    }

    canvas, axes, mark = tree.draw(
        width=400,
        height=600,
        **mystyle,
    )

    toyplot.pdf.render(canvas, out)

if __name__ == "__main__":
    main()



#!/usr/bin/env python

# Standard library imports
import argparse
import sys

import numpy as np
import pandas as pd
import scipy.stats as stats
from sklearn_genetic.space import Continuous, Categorical, Integer

from utils.misc import get_processor_name
from utils.misc import generate_012_genotypes

# Custom module imports
from snpio import GenotypeData
from read_input.simgenodata import SimGenotypeData
from impute.estimators import *
from impute.simple_imputers import *


def main():
    """[Class instantiations and main package body]"""

    args = get_arguments()

    if args.str and args.phylip:
        sys.exit("Error: Only one file type can be specified")

        # If VCF file is specified.
    if args.str:
        if not args.pop_ids and args.popmap is None:
            raise TypeError("Either --pop_ids or --popmap must be specified\n")

        if args.pop_ids:
            print("\n--pop_ids was specified as column 2\n")
        else:
            print(
                "\n--pop_ids was not specified; "
                "using popmap file to get population IDs\n"
            )

        if args.onerow_perind:
            print("\nUsing one row per individual...\n")
        else:
            print("\nUsing two rows per individual...\n")

        if args.onerow_perind:
            data = GenotypeData(
                filename=args.str,
                filetype="structure1row",
                popmapfile=args.popmap,
                guidetree=args.treefile,
                qmatrix_iqtree=args.iqtree,
            )
        else:
            data = GenotypeData(
                filename=args.str,
                filetype="structure2row",
                popmapfile=args.popmap,
                guidetree=args.treefile,
                qmatrix_iqtree=args.iqtree,
            )

    if args.phylip:
        if args.pop_ids or args.onerow_perind:
            print(
                "\nPhylip file was used with structure arguments; ignoring "
                "structure file arguments\n"
            )

        if args.popmap is None:
            raise TypeError("No popmap file supplied with PHYLIP file\n")

        data = GenotypeData(
            filename=args.phylip,
            filetype="phylip",
            popmapfile=args.popmap,
            guidetree=args.treefile,
            qmatrix_iqtree=args.iqtree,
            siterates_iqtree=args.rates,
        )

        prefix = "c0.001_s0.009_gtrgamma_i0.0"
        sim = SimGenotypeData(data, prop_missing=0.1, strategy="random")

        nmf = ImputeNMF(genotype_data=sim)

        accuracy = sim.accuracy(nmf)
        print("Accuracy:", accuracy)

        phylo = ImputePhylo(genotype_data=sim, save_plots=False)

        accuracy = sim.accuracy(phylo)
        print("Accuracy:", accuracy)

        nlpca = ImputeNLPCA(
            genotype_data=sim, initial_strategy="populations", cv=5
        )
        accuracy = sim.accuracy(nlpca)
        print("Accuracy:", accuracy)

        ubp = ImputeUBP(genotype_data=sim, initial_strategy="populations")
        accuracy = sim.accuracy(ubp)
        print("Accuracy:", accuracy)

        # vae = ImputeVAE(
        #     genotype_data=sim,
        #     initial_strategy="populations"
        # )
        # accuracy = sim.accuracy(vae)
        # print("Accuracy:",accuracy)


def get_arguments():
    """[Parse command-line arguments. Imported with argparse]

    Returns:
        [argparse object]: [contains command-line arguments; accessed as method]
    """

    parser = argparse.ArgumentParser(
        description="Simulate missing data on GenotypeData object",
        add_help=False,
    )

    required_args = parser.add_argument_group("Required arguments")
    filetype_args = parser.add_argument_group(
        "File type arguments (choose only one)"
    )
    structure_args = parser.add_argument_group("Structure file arguments")
    optional_args = parser.add_argument_group("Optional arguments")

    # File Type arguments
    filetype_args.add_argument(
        "-s", "--str", type=str, required=False, help="Input structure file"
    )
    filetype_args.add_argument(
        "-p", "--phylip", type=str, required=False, help="Input phylip file"
    )

    filetype_args.add_argument(
        "-t",
        "--treefile",
        type=str,
        required=False,
        default=None,
        help="Newick-formatted treefile",
    )

    # Structure Arguments
    structure_args.add_argument(
        "--onerow_perind",
        default=False,
        action="store_true",
        help="Toggles on one row per individual option in structure file",
    )
    structure_args.add_argument(
        "--pop_ids",
        default=False,
        required=False,
        action="store_true",
        help="Toggles on population ID column (2nd col) in structure file",
    )

    ## Optional Arguments
    optional_args.add_argument(
        "-m",
        "--popmap",
        type=str,
        required=False,
        default=None,
        help="Two-column tab-separated population map file: inds\tpops. No header line",
    )
    optional_args.add_argument(
        "-i",
        "--iqtree",
        type=str,
        required=False,
        help=".iqtree output file containing Rate Matrix Q",
    )

    optional_args.add_argument(
        "-r",
        "--rates",
        type=str,
        required=False,
        help="IQ-TREE site-rates output file",
    )

    optional_args.add_argument(
        "--prefix",
        type=str,
        required=False,
        default="output",
        help="Prefix for output files",
    )

    # Add help menu
    optional_args.add_argument(
        "-h", "--help", action="help", help="Displays this help menu"
    )

    # If no command-line arguments are called then exit and call help menu.
    if len(sys.argv) == 1:
        print("\nExiting because no command-line options were called.\n")
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    main()
